"""RF calibrations like virtual probe, RF actuator offset/imbalance, forward and reflected, and power calibrations."""
#############################################################################
#  Copyright (c) 2023 by Paul Scherrer Institute, Switzerland
#  All rights reserved.
#  Authors: Zheqiao Geng
#############################################################################
'''
#########################################################################
Here collects routines for RF system calibration

Implemented:    
    - calib_vprobe       : calib. cavity virtual probe with forward/reflected signals
    - calib_dac_offs     : calib. DAC offset with I/Q modulator direct upconversion
    - calib_iqmod        : calib. I/Q modulator imbalance with direct upconversion
    - calib_cav_for      : calib. cavity forward signal using forward and probe
    - calib_ncav_for_ref : calib. cavity forward/reflected signals with constant QL and detuning
    - calib_scav_for_ref : calib. cavity forward/reflected signals with time-varying QL and detuning
    - for_ref_volt2power : calib. forward/reflected power from forward/reflected voltage
    - phasing_energy     : calib. energy gain and beam phase with phase scan and energy meas.
    - egain_cav_power    : est. steady-state standing-wave cavity voltage from drive power
    - egain_cgstr_power  : est. const gradient traveling-wave structure ACC voltage from drive power
    - calib_vsum_poor    : calib. vector sum by rotating and scaling referring to one channel
    - calib_sys_gain_pha : calib. system gain and system phase

To be implemented: 
    - calibrate the Vacc and phi_b of each cavity using beam transient
    - calibrate the vector sum (beam transient based)

Some algorithms are referred to the following books:
S. Simrock and Z. Geng, Low-Level Radio Frequency Systems, Springer, 2022
https://link.springer.com/book/10.1007/978-3-030-94419-3 ("LLRF Book")
#########################################################################
'''
import numpy as np

from llrflibs.rf_sysid import *
from llrflibs.rf_fit import *

def calib_vprobe(vc, vf_m, vr_m):
    '''
    Calibrate the virtual probe signal of the cavity with ``vc = m * vf_m + n * vr_m``
    aiming at reconstructing the probe signal with the forward and reflected
    signals. This is useful when the probe is temperally not available.
    
    Parameters:
         vc:   numpy array (complex), cavity probe waveform (as reference plane)
         vf_m: numpy array (complex), cavity forward waveform (meas.)
         vr_m: numpy array (complex), cavity reflected waveform (meas.)
         
    Returns:
        status: boolean, success (True) or fail (False)
        m, n:   float (complex), calibration coefficients    
    
    Note: 
        The waveforms should have been aligned in time, i.e., the relative delays between them should have been removed.
    '''
    # check the input
    if (not (vc.shape == vf_m.shape == vr_m.shape)) or (len(vc) < 3):
        return False, None, None

    # formulate the problem of Ax = B
    X = np.linalg.lstsq(np.vstack((vf_m, vr_m)).T, vc, rcond = None)

    # return the status, m, n
    return True, X[0][0], X[0][1]

def calib_dac_offs(vdac, viqm, sig_ids, sig_ide, leak_ids, leak_ide, delay = 0):
    '''
    Calibrate the DAC offset to remove carrier leakage for direct upconversion.
    
    Refer to LLRF Book section 9.2.2.
    
    Parameters:
        vdac:     numpy array (complex), DAC output waveform
        viqm:     numpy array (complex), I/Q modulator output meas. waveform
        sig_ids:  int, starting index of signals
        sig_ide:  int, ending index of signals
        leak_ids: int, starting index of leakage
        leak_ide: int, ending index of leakage
        delay :   int, delay in number of points between I/Q out and DAC WFs,
                        it is needed to be set reflecting the actual delay for
                        better results
    Returns:
        status:   boolean, success (True) or fail (False)
        offset_I: float, I offset to be added to the actual DAC output
        offset_Q: float, Q offset to be added to the actual DAC output
    '''
    # check the input
    N = vdac.shape[0]
    if vdac.shape != viqm.shape or \
       sig_ids  < 0 or sig_ide  < 0 or \
       leak_ids < 0 or leak_ide < 0 or \
       (sig_ids  + delay) >= N or (sig_ide  + delay) >= N or \
       (leak_ids + delay) >= N or (leak_ide + delay) >= N or \
       sig_ids >= sig_ide or leak_ids >= leak_ide:
        return False, None, None

    # make the calculation of the offset
    vdac_sig  = np.mean(vdac[sig_ids  : sig_ide])
    vdac_leak = np.mean(vdac[leak_ids : leak_ide])
    viqm_sig  = np.mean(viqm[sig_ids  + delay : sig_ide  + delay])
    viqm_leak = np.mean(viqm[leak_ids + delay : leak_ide + delay])

    voff = -(vdac_sig - vdac_leak) / (viqm_sig - viqm_leak) * viqm_leak
    
    return True, np.real(voff), np.imag(voff)

def calib_iqmod(vdac, viqm):
    '''
    Calibrate the I/Q modulator imbalance, return ``inv(M)`` with 
    ``[real(viqm) imag(viqm)]' = M * [real(vdac) imag(vdac)]' + [iqm_offs_I iqm_offs_Q]'``
    This method is OK to calibrate the imbalance matrix M (use its inverse
    to pre-distort the DAC output) but the offset is not accurate to determine
    the DAC offset calibration. It is suggested to calibrate the DAC offset
    first using the routine ``calib_dac_offs`` before executing this function.
    
    Refer to paper "Geng Z, Hong B (2016) Design and calibration of an RF actuator for 
    low-level RF systems. IEEE Trans Nucl Sci 63(1):281-287".
 
    Parameters:
        vdac: numpy array (complex), DAC actuation points
        viqm: numpy array (complex), I/Q modulator meas. points
        
    Returns:
        status: boolean, success (True) or fail (False)
        invM:   numpy matrix, inversion of M, can be directly used to 
                    correct the I/Q modulator imbalance
        viqm_s: numpy array (complex), scaled I/Q modulator output
        viqm_c: numpy array (complex), re-constructed I/Q mod. output
    '''
    # check the input
    if vdac.shape != viqm.shape or vdac.shape[0] < 3:
        return (False,) + (None,)*3

    # scale & rotate the I/Q modulator output to align with DAC
    viqm_s = np.mean(vdac/viqm) * viqm

    # calibrate with least-square method
    Y = np.matrix(np.vstack((np.real(viqm_s), np.imag(viqm_s))))
    U = np.matrix(np.vstack((np.real(vdac), np.imag(vdac), np.ones(vdac.shape[0]))))
    M = Y * U.T * np.linalg.inv(U * U.T)

    # reconstruct the scaled output
    Y_t = np.array(M * U)
    viqm_c = Y_t[0] + 1j*Y_t[1]

    return True, np.linalg.inv(M[:2, :2]), viqm_s, viqm_c

def calib_cav_for(vc, vf_m, pul_ids, pul_ide, half_bw, detuning, Ts, beta = 1e4):
    '''
    Calibrate the cavity forward signal: ``vf = a * vf_m``. The resulting ``vf`` is
    the cavity drive signal at the same reference plane as the cavity probe ``vc``.
    
    Refer to LLRF Book section 4.6.1.3.
    
    Parameters:
        vc:       numpy array (complex), cavity probe waveform (reference plane)
        vf_m:     numpy array (complex), cavity forward waveform (meas.)
        pul_ids:  int, starting index for calculation (calculation window should 
                   cover the entire pulse for NC cavities; use the part close to 
                   pulse end for SC cavities)
        pul_ide:  int, ending index for calculation 
        half_bw:  float, half bandwidth of the cavity, rad/s
        detuning: float, detuning of the cavity, rad/s
        Ts:       float, sampling time, s
        beta:     float, input coupling factor (needed for NC cavities; 
                   for SC cavities, can use the default value, or you can 
                   specify it if more accurate calibration is needed)
                   
    Returns:
        status:   boolean, success (True) or fail (False)
        a:        float (complex), calibrate coefficient
    
    Note: 
        The waveforms should have been aligned in time, i.e., the relative delays between them should have been removed.
    '''
    # check the input
    if (not (vc.shape == vf_m.shape)) or (len(vc) < 3) or \
       (pul_ids < 0) or (pul_ide <= pul_ids) or (Ts <= 0) or \
       (half_bw <= 0):
        return False, None

    # estimate the theoritical forward from probe
    status, vf_est, _ = cav_drv_est(vc, half_bw, Ts, detuning, beta)
    if not status:
        return False, None

    # make the calibration
    return True, np.mean(vf_est[pul_ids:pul_ide]) / np.mean(vf_m[pul_ids:pul_ide])

def calib_ncav_for_ref(vc, vf_m, vr_m, pul_ids, pul_ide, half_bw, detuning, Ts, beta = 1.0):
    '''
    Calibrate the cavity forward and reflected signals for a normal conducting 
    cavity with constant loaded Q and detuning:
    ``vf = a * vf_m + b * vr_m, vr = c * vf_m + d * vr_m``
    The resulting ``vf`` and ``vr`` are the cavity drive/reflected signals at the 
    same reference plane as the cavity probe ``vc``.
    
    Refer to LLRF Book section 9.3.5.
    
    Parameters:
        vc:        numpy array (complex), cavity probe waveform (reference plane)
        vf_m:      numpy array (complex), cavity forward waveform (meas.)
        vr_m:      numpy array (complex), cavity reflected waveform (meas.)
        pul_ids:   int, starting index for calculation (calculation window should 
                    cover entire pulse)
        pul_ide:   int, ending index for calculation
        half_bw:   float, half bandwidth of the cavity, rad/s
        detuning:  float, detuning of the cavity, rad/s
        Ts:        float, sampling time, s
        beta:      float, input coupling factor (needed for NC cavities)
        
    Returns:
        status:    boolean, success (True) or fail (False)
        a,b,c,d:   float (complex), calibrate coefficients
    
    Note: 
        The waveforms should have been aligned in time, i.e., the relative delays between them should have been removed.
    '''    
    # check the input
    if (not (vc.shape == vf_m.shape == vr_m.shape)) or (len(vc) < 3) or \
       (pul_ids < 0) or (pul_ide <= pul_ids) or (Ts <= 0) or \
       (half_bw <= 0):
        return (False,) + (None,)*4

    # estimate the theoritical forward and reflected signals from probe
    status, vf_est, vr_est = cav_drv_est(vc, half_bw, Ts, detuning, beta)
    if not status:
        return (False,) + (None,)*4

    # calibrate the virtual probe to get: vc = m * vf_m + n * vr_m
    # with m = a + c, n = b + d
    status, m, n = calib_vprobe(vc, vf_m, vr_m)
    if not status:
        return (False,) + (None,)*4

    # use "calib_vprobe" which solves a multi-linear fitting problem
    status, a, b = calib_vprobe(vf_est[pul_ids:pul_ide], 
                                vf_m[pul_ids:pul_ide], 
                                vr_m[pul_ids:pul_ide])
    if not status:
        return (False,) + (None,)*4

    # finally return the results
    return True, a, b, m - a, n - b

def calib_scav_for_ref(vc, vf_m, vr_m, pul_ids, pul_ide, decay_ids, decay_ide, 
                       half_bw, detuning, Ts, beta = 1e4):
    '''
    Calibrate the cavity forward and reflected signals for a superconducting 
    cavity with time-varying loaded Q or detuning:
    ``vf = a * vf_m + b * vr_m, vr = c * vf_m + d * vr_m``
    The resulting ``vf`` and ``vr`` are the cavity drive/reflected signals at the 
    same reference plane as the cavity probe ``vc``.
    
    Refer to LLRF Book section 9.3.5.
    
    Parameters:
        vc:        numpy array (complex), cavity probe waveform (reference plane)
        vf_m:      numpy array (complex), cavity forward waveform (meas.)
        vr_m:      numpy array (complex), cavity reflected waveform (meas.)
        pul_ids:   int, starting index for calculation (calculation window should 
                    take the pulse end part close to decay)
        pul_ide:   int, ending index for calculation
        decay_ids: int, starting id of calculation window at decay stage
        decay_ide: int, ending id of calculation window at decay stage
        half_bw:   float, half bandwidth of the cavity (derived from early part of decay), rad/s
        detuning:  float, detuning of the cavity (derived from early part of decay), rad/s
        Ts:        float, sampling time, s
        beta:      float, input coupling factor (for SC cavities, can use the default value,
                    or you can specify it if more accurate calibration is needed)
                    
    Returns:
        status:    boolean, success (True) or fail (False)
        a,b,c,d:   float (complex), calibrate coefficients
    
    Note: 
        The waveforms should have been aligned in time, i.e., the relative delays between them should have been removed.
    '''
    # check the input
    if (not (vc.shape == vf_m.shape == vr_m.shape)) or (len(vc) < 3) or \
       (pul_ids < 0) or (pul_ide <= pul_ids) or (Ts <= 0) or \
       (decay_ids < 0) or (decay_ide <= decay_ids) or (half_bw <= 0):
        return (False,) + (None,)*4

    # estimate the theoritical forward and reflected signals from probe
    status, vf_est, vr_est = cav_drv_est(vc, half_bw, Ts, detuning, beta)
    if not status:
        return (False,) + (None,)*4

    # calibrate the virtual probe to get: vc = m * vf_m + n * vr_m
    # with m = a + c, n = b + d
    status, m, n = calib_vprobe(vc, vf_m, vr_m)
    if not status:
        return (False,) + (None,)*4        

    # find from decay z = b/a
    z = -np.mean(vf_m[decay_ids:decay_ide]) / np.mean(vr_m[decay_ids:decay_ide])

    # find a good "a" to match the estimated forward signal close to pulse end
    #   vf_est = a * vf_m + a * z * vr_m
    a = np.mean(vf_est[pul_ids:pul_ide]) / np.mean(vf_m[pul_ids:pul_ide] + z * vr_m[pul_ids:pul_ide])
    
    # get other parameters
    b = a * z
    c = m - a
    d = n - b

    # finally return the results
    return True, a, b, c, d

def for_ref_volt2power(roQ_or_RoQ, QL, 
                       vf_pcal  = None, 
                       vr_pcal  = None, 
                       beta     = 1e4, 
                       machine  = 'linac'):
    '''
    Convert the calibrated forward and reflected signals (in physical unit, V)
    to forward and reflected power (in W).
    
    Refer to LLRF Book section 3.3.9.
    
    Parameters:
        roQ_or_RoQ:  float, cavity r/Q of Linac or R/Q of circular accelerator, Ohm
                      (see the note below)
        QL:          float, loaded quality factor of the cavity
        vf_pcal:     numpy array (complex), forward waveform (calibrated to physical unit)
        vf_pcal:     numpy array (complex), reflected waveform (calibrated to physical unit)
        beta:        float, input coupling factor (needed for NC cavities; 
                      for SC cavities, can use the default value, or you can 
                      specify it if more accurate result is needed)
        machine:     string, 'linac' or 'circular', used to select r/Q or R/Q
        
    Returns:
        status:      boolean, success (True) or fail (False)
        for_power:   numpy array, waveform of forward power (if input is not None), W
        ref_power:   numpy array, waveform of reflected power (if input is not None), W
        C:           float (complex), calibration coefficient: power_W = C * Volt_V^2
    
    Note: 
          Linacs define the ``r/Q = Vacc**2 / (w0 * U)`` while circular machines 
          define ``R/Q = Vacc**2 / (2 * w0 * U)``, where ``Vacc`` is the accelerating
          voltage, ``w0`` is the angular cavity resonance frequency and ``U`` is the
          cavity energy storage. Therefore, to use this function, one needs to 
          specify the ``machine`` to be ``linac`` or ``circular``. Generally, we have
          ``R/Q = 1/2 * r/Q``.
    '''
    # check the input
    if (roQ_or_RoQ <= 0.0) or (QL <= 0.0) or (beta <= 0.0):
        return (False,) + (None,)*3

    # calculate the loaded resistance
    if machine == 'circular':
        RL = roQ_or_RoQ * QL
    else:
        RL = 0.5 * roQ_or_RoQ * QL

    # calculate the coefficient to convert voltage to power
    C = beta / (beta + 1) / (2 * RL)

    # convert the voltage to power if they are valid
    for_power = ref_power = None
    if vf_pcal is not None: for_power = C * np.abs(vf_pcal)**2
    if vr_pcal is not None: ref_power = C * np.abs(vr_pcal)**2

    return True, for_power, ref_power, C

def phasing_energy(phi_vec, E_vec, machine  = 'linac'):
    '''
    Calibrate the beam phase using the beam energy measured by scanning RF phase.
    
    Refer to LLRF Book section 9.3.2.3.
    
    Parameters:
        phi_vec: numpy array, phase measurement, degree
        E_vec:   numpy array, beam energy measurement
        machine: string, 'linac' or 'circular', see the note below
        
    Returns:
        status:  boolean, success (True) or fail (False)
        Egain:   float, maximum energy gain of the RF station
        phi_off: float, phase offset that should be added to the phase meas, deg
    
    Note: 
          Circular accelerator use sine as convention of beam phase definition,
          resulting in 90 degree for on-crest acceleration; while for Linacs, 
          the consine convention is used with 0 degree for on-crest acceleration.
    '''
    # check the input
    if (not phi_vec.shape == E_vec.shape) or (phi_vec.shape[0] < 3):
        return False, None, None

    # make the fitting
    if machine == 'circular':
        status, Egain, phi_rad, c = fit_sincos(phi_vec*np.pi/180.0, E_vec, target = 'sin')
    else:
        status, Egain, phi_rad, c = fit_sincos(phi_vec*np.pi/180.0, E_vec, target = 'cos')

    if not status:
        return False, None, None

    # return the energy gain and phase offset
    return True, Egain, phi_rad*180.0/np.pi

def egain_cav_power(pfor, roQ_or_RoQ, QL, beta = 1e4, machine = 'linac'):
    '''
    Standing-wave cavity energy gain estimate from RF drive power without beam loading.
    
    Refer to LLRF Book section 9.3.2.1, equation (9.13).
    
    Parameters:
        pfor:        float, cavity drive power, W
        roQ_or_RoQ:  float, cavity r/Q of Linac or R/Q of circular accelerator, Ohm
                      (see the note below)
        QL:          float, loaded quality factor of the cavity
        beta:        float, input coupling factor (needed for NC cavities; 
                      for SC cavities, can use the default value, or you can 
                      specify it if more accurate result is needed)
        machine:     string, ``linac`` or ``circular``, used to select r/Q or R/Q 
    Returns:
        status:      boolean, success (True) or fail (False)
        vc0:         float, desired cavity voltage for the given drive power
    
    Note: 
          Linacs define the ``r/Q = Vacc**2 / (w0 * U)`` while circular machines 
          define ``R/Q = Vacc**2 / (2 * w0 * U)``, where ``Vacc`` is the accelerating
          voltage, ``w0`` is the angular cavity resonance frequency and ``U`` is the
          cavity energy storage. Therefore, to use this function, one needs to 
          specify the ``machine`` to be ``linac`` or ``circular``. Generally, we have
          ``R/Q = 1/2 * r/Q``.
    '''
    # check the input
    if (pfor < 0.0) or (roQ_or_RoQ <= 0.0) or (QL <= 0.0) or (beta <= 0.0):
        return False, None

    # calculate the loaded resistance
    if machine == 'circular':
        RL = roQ_or_RoQ * QL
    else:
        RL = 0.5 * roQ_or_RoQ * QL

    # calculate the maximum energy gain (accelerating voltage)
    vc0 = 2.0 * np.sqrt(2.0 * beta * RL * pfor / (beta + 1.0))   
    return True, vc0

def egain_cgstr_power(pfor, f0, rs, L, Q, Tf):
    '''
    Traveling-wave constant gradient structure energy gain estimate from RF 
    drive power without beam loading.
    
    Refer to LLRF Book section 9.3.2.1, equation (9.14).
    
    Parameters:
        pfor: float, structure input RF power, W
        f0:   float, RF operating frequency, Hz
        rs:   float, shunt impedance per unit length, Ohm/m
        L:    float, length of the traveling-wave structure, m
        Q:    float, quality factor of the TW strcuture
        Tf:   float, filling time of the TW structure, s

    Returns:
        status: boolean, success (True) or fail (False)
        vc0:    float, desired accelerating voltage for the given drive power
    '''
    # check the input
    if (f0 <= 0.0) or (pfor < 0.0) or (rs <= 0.0) or (L <= 0.0) or \
       (Q <= 0.0) or (Tf <= 0.0):
        return False, None

    # calculate the total power attenuation factor
    tao_tw = Tf * np.pi * f0 / Q

    # calculate the maximum energy gain (accelerating voltage)
    vacc0 = np.sqrt(L * pfor * rs * (1.0 - np.exp(-2.0 * tao_tw)))
    return True, vacc0

def calib_vsum_poor(amp_vec, pha_vec, ref_ch = 0):
    '''
    Calibrate the vector sum (poor man's solution by rotating and scaling all other 
    channels refering to the first channel).
    
    Parameters:
        amp_vec: numpy array, amplitude of all channels
        pha_vec: numpy array, phase of all channels, deg
        ref_ch:  int, reference channel, between 0 and (num of channels - 1)
        
    Returns:
        status:  boolean, success (True) or fail (False)
        scale:   numpy array, scale factor for all channels
        phase:   numpy array, phase offset adding to all channels, deg
    '''
    # check the input
    if (not amp_vec.shape == pha_vec.shape) or \
       (ref_ch < 0 or ref_ch >= amp_vec.shape[0]):
        return False, None, None

    # make the calculation
    scale = amp_vec[ref_ch] / amp_vec
    phase = pha_vec[ref_ch] - pha_vec

    return True, scale, phase

def calib_sys_gain_pha(vact, vf, beta = 1e4):
    '''
    Calibrate the system gain and sytem phase.
    
    Refer to LLRF Book section 9.4.2.
    
    Parameters:
        vact:      numpy array (complex), RF actuation waveform (usually DAC output)
        vf:        numpy array (complex), cavity forward signal (calibrated to 
                    the reference plane of the cavity voltage or vector-sum voltage)
        beta:      float, input coupling factor (needed for NC cavities; 
                    for SC cavities, can use the default value, or you can 
                    specify it if more accurate result is needed)    
    Returns:
        status:    boolean, success (True) or fail (False)
        sys_gain:  numpy array, waveform of system gain
        sys_phase: numpy array, waveform of system phase, deg
    '''
    # check the input
    if (not vact.shape == vf.shape) or (beta <= 0.0):
        return False, None, None

    # make the calibration
    G         = 2.0 * beta / (beta + 1.0) * vf / vact
    sys_gain  =  np.abs(G)
    sys_phase = -np.angle(G, deg = True)

    # return the results
    return True, sys_gain, sys_phase






















"""Analyze, generate and filter noise."""
#############################################################################
#  Copyright (c) 2023 by Paul Scherrer Institute, Switzerland
#  All rights reserved.
#  Authors: Zheqiao Geng
#############################################################################
'''
#########################################################################
Here collects routines for RF noise analysis
 
Implemented:
    - calc_psd_coherent     : calculate the power spectral density (PSD) of a 
                              coherent sampled data series
    - calc_psd              : calculate the PSD of a general data series
    - rand_unif             : generate uniform distributed random numbers
    - gen_noise_from_psd    : generate noise series from a given PSD spectrum
    - calc_rms_from_psd     : calculate the RMS jitter from a given PSD spectrum
    - notch_filt            : apply notch filter to a data series
    - design_notch_filter   : design a notch filter in state-space format
    - filt_step             : apply a single time step of state-space filter
    - rand_sine             : generate random sine signals
    - gen_rand_sine_from_psd: generate random sine functions from DSB PSD
    - moving_avg            : moving average with group delay compensated

To be implemented:
    - correlation
    - sum and sub of dB values
    - random sub/sum (RMS values)
    - estimate the RF detector nonlinearity caused meas. errors

Some algorithms are referred to the following books:
S. Simrock and Z. Geng, Low-Level Radio Frequency Systems, Springer, 2022
https://link.springer.com/book/10.1007/978-3-030-94419-3 ("LLRF Book")
#########################################################################
'''
import numpy as np
from scipy import signal
from scipy.ndimage import uniform_filter1d

def calc_psd_coherent(data, fs, bit = 0, n_noniq = 1, plot = False):
    '''
    Calculate the power spectral density of the input waveform (coherent sampling).
    
    Refer to LLRF Book section 6.1.2.

    Parameters:
        data:        numpy array, 1-D array of the raw waveform
        fs:          float, sampling frequency, Hz
        bit:         int, number of bit of data. If bit = 0, do not scale data
        n_noniq:     int, non-I/Q parameters (n samples cover a full cycle)
        plot:        boolean, True for plot the spectrum
        
    Returns:
        freq:        numpy array, frequency waveform, Hz
        amp_resp:    numpy array, amplitude response, dBFS/Hz
        pha_resp:    numpy array, phase response, degree
        signal_freq: float, signal frequency, Hz
        signal_mag:  float, signal level, dBFS
        noise_flr:   float, noise floor, dBFS/Hz
        snr:         float, signal-to-noise ratio, dB
        bin_db:      float, offset for an FFT bin freq space, dB
        status:      boolean, success (True) or fail (False)

    Note:
        The unit dBFS is for ADC raw data, if for phase in radian, it should be
        replaced by dBrad^2; if for relative amplitude, should be replaced by dB.
    '''
    # results
    result = {'status': False}

    # get the input data size
    if isinstance(data, np.ndarray): L = data.shape[0]
    elif isinstance(data, list):     L = len(data)
    else:                            L = 0

    # tailor the data: multiple of n_noniq (coherent)
    if n_noniq <= 0: n_noniq = 1                        # avoid dividing 0
    N = int(L / n_noniq) * n_noniq

    # check the input
    if N < 3 or fs <= 0 or bit < 0:
        return result

    # calculate the FFT and get the results in Nyquist band
    if bit > 0:
        fftd = data[0:N] / (2.0**(bit-1)/np.sqrt(2.0))  # tailor the data and normalize with full scale
    else:
        fftd = data[0:N]

    Y    = np.fft.fft(fftd)                             # do FFT
    freq = np.arange(N) / N * fs                        # frequency vector, Hz
    ids  = freq <= 0.5 * fs                             # index of Nyquist band
    Y    = Y[ids]                                       # tailor the response
    freq = freq[ids]                                    # tailor the frequency vector

    # calculate the PSD
    PSD  = np.abs(Y) * np.abs(Y) / (N * fs)             # SSB spectra, linear scale relative power / Hz
    if (N % 2) == 0: PSD[1:-1] *= 2.0                   # convert to DSB spectra (exclude DC and f_nyquist)
    else:            PSD[1:]   *= 2.0                   # convert to DSB spectra (exclude DC)

    # calculate the derived quantities
    amp_resp    = 10.0 * np.log10(PSD)                  # in dBFS/Hz
    pha_resp    = np.angle(Y, deg = True)               # in degree
    bin_dB      = 10.0 * np.log10(fs / N)               # offset in dB for an FFT bin frequency band

    sig_id      = np.argmax(amp_resp)                   # bin index of the signal
    sig_level   = amp_resp[sig_id] + bin_dB             # signal power level, dBFS

    est_flr     = np.mean(amp_resp)                     # average the amplitude response in dB, a guess of noise floor
    spur_id     = amp_resp > est_flr + 15               # find any spur 15 dB higher than the noise floor

    noise_bins  = PSD[np.invert(spur_id)]               # find all the noise bins (linear scale)
    noise_flr   = 10.0 * np.log10(np.mean(noise_bins))  # average the noise spectral density, dBFS/Hz

    noise_level = noise_flr + 10.0 * np.log10(fs/2)     # noise power level, dBFS
    snr         = sig_level - noise_level               # SNR, dB

    # collect the results
    result['freq']        = freq                        # Hz
    result['psd']         = PSD                         # linear PSD (power / Hz)
    result['amp_resp']    = amp_resp                    # dBFS/Hz
    result['pha_resp']    = pha_resp                    # degree
    result['signal_freq'] = freq[sig_id]                # signal frequency, Hz
    result['signal_mag']  = sig_level                   # signal magnitude, dBFS
    result['noise_flr']   = noise_flr                   # dBFS/Hz
    result['snr']         = snr                         # dB
    result['bin_db']      = bin_dB                      # dB
    result['status']      = True

    # make the plot
    if plot:
        from llrflibs.rf_plot import plot_calc_psd
        plot_calc_psd(result)
        
    # return
    return result

def calc_psd(data, fs, bit = 0, plot = False):
    '''
    Calculate the power spectral density of the input waveform (nominal sampling).
    
    Refer to LLRF Book section 6.1.2.

    Parameters:
        data:        numpy array, 1-D array of the raw waveform
        fs:          float, sampling frequency, Hz
        bit:         int, number of bit of data. If bit = 0, do not scale data
        plot:        boolean, True for plot the spectrum
        
    Returns:
        freq:        numpy array, frequency waveform, Hz
        amp_resp:    numpy array, amplitude response, dBFS/Hz
        pha_resp:    numpy array, phase response, degree
        signal_freq: float, signal frequency, Hz
        signal_mag:  float, signal level, dBFS
        noise_flr:   float, noise floor, dBFS/Hz
        snr:         float, signal-to-noise ratio, dB
        bin_db:      float, offset for an FFT bin freq space, dB
        enbw_db:     float, offset for an FFT bin freq space (correct windowing), dB
        status:      boolean, success (True) or fail (False)

    Note:
        The unit dBFS is for ADC raw data, if for phase in radian, it should be
        replaced by dBrad^2; if for relative amplitude, should be replaced by dB
    '''
    # results
    result = {'status': False}

    # get the input data size
    if isinstance(data, np.ndarray): 
        N = data.shape[0]
    elif isinstance(data, list):     
        N = len(data)
        data = np.array(data)
    else:
        N = 0

    # check the input
    if N < 3 or fs <= 0 or bit < 0:
        return result

    # calculate the FFT and get the results in Nyquist band
    if bit > 0:
        fftd = data[0:N] / (2.0**(bit-1)/np.sqrt(2.0))  # tailor the data and normalize with full scale
    else:
        fftd = data[0:N]

    win  = np.blackman(N)                               # blackman window    
    Y    = np.fft.fft(fftd * win)                       # do FFT on windowed data
    freq = np.arange(N) / N * fs                        # frequency vector, Hz
    ids  = freq <= 0.5 * fs                             # index of Nyquist band
    Y    = Y[ids]                                       # tailor the response
    freq = freq[ids]                                    # tailor the frequency vector

    Wn   = 1.0 / N * np.sum(win * win)                  # window correction factor
    Ws   = np.mean(win)**2 
    ENBW = Wn / Ws * fs / N                             # effective noise bandwidth for calculating the signal level

    # calculate the PSD
    PSD  = np.abs(Y) * np.abs(Y) / (N * fs * Wn)        # SSB spectra, linear scale relative power / Hz
    if (N % 2) == 0: PSD[1:-1] *= 2.0                   # convert to DSB spectra (exclude DC and f_nyquist)
    else:            PSD[1:]   *= 2.0                   # convert to DSB spectra (exclude DC)

    # calculate the derived quantities
    amp_resp    = 10.0 * np.log10(PSD)                  # in dBFS/Hz
    pha_resp    = np.angle(Y, deg = True)               # in degree
    bin_dB      = 10.0 * np.log10(fs / N)               # offset in dB for an FFT bin frequency band
    enbw_dB     = 10.0 * np.log10(ENBW)                 # offset in dB for an FFT bin frequency band (corrected the windowing)
                                                        # Note: "bin_dB" is used to integrate the noise power and the signal power
                                                        #       (using all signal bins including the leaked ones); "enbw_dB" is only
                                                        #       used in calcuating the signal level by taking only one central bin 
                                                        #       of the signal - to avoid integrating the spreaded signal bins (see
                                                        #       below for the "sig_level" calculation)

    sig_id      = np.argmax(amp_resp)                   # bin index of the signal
    sig_level   = amp_resp[sig_id] + enbw_dB            # signal power level (enbw_dB corrects the signal spectra leakage), dBFS
                                                        # Note: by this treatment, we obtain a "sig_level" equivalent integrating
                                                        #       all the signal bins that are formed by the leakage of the spectrum

    est_flr     = np.mean(amp_resp)                     # average the amplitude response in dB, a guess of noise floor
    spur_id     = amp_resp > est_flr + 15               # find any spur 15 dB higher than the noise floor

    noise_id    = np.invert(spur_id)                    # indices of noise bins
    noise_bins  = PSD[noise_id]                         # find all the noise bins (linear scale)    
    noise_flr   = 10.0 * np.log10(np.mean(noise_bins))  # average the noise spectral density, dBFS/Hz

    noise_level = noise_flr + 10.0 * np.log10(fs/2)     # noise power level, dBFS
    snr         = sig_level - noise_level               # SNR, dB

    # collect the results
    result['freq']        = freq                        # Hz
    result['psd']         = PSD                         # complex PSD
    result['amp_resp']    = amp_resp                    # dBFS/Hz
    result['pha_resp']    = pha_resp                    # degree
    result['signal_freq'] = freq[sig_id]                # signal frequency, Hz
    result['signal_mag']  = sig_level                   # signal magnitude, dBFS
    result['noise_flr']   = noise_flr                   # dBFS/Hz
    result['snr']         = snr                         # dB
    result['bin_db']      = bin_dB                      # dB
    result['enbw_db']     = enbw_dB                     # dB
    result['status']      = True

    # make the plot
    if plot:
        from llrflibs.rf_plot import plot_calc_psd
        plot_calc_psd(result)

    # return
    return result

def rand_unif(low = 0.0, high = 1.0, n = 1):
    '''
    produce random number within a certain range.
    
    Parameters:
        n:    int, number of output
        low:  float, low limit of the data
        high: float, high limit of the data
        
    Returns:
        val:  if n = 1, it is a float number, if n > 1, it is a np array
    '''
    # check the input
    if n < 1: n = 1

    # generate the random numbers
    val = np.random.uniform(low, high, n)

    # make return
    if n == 1: return val[0]
    else:      return val

def gen_noise_from_psd(freq_vector, pn_vector, fs, N):
    '''
    generate noise series from DSB PSD.

    Parameters:
        freq_vector: numpy array, offset frequency from carrier, Hz
        pn_vector:   numpy array, DSB noise PSD, dBrad^2/Hz for phase noise
        fs:          float, sampling frequency, Hz
        N:           float, number of samples
        
    Returns:
        status:      boolean, success (True) or fail (False)
        pn_series:   numpy array, time series of phase/amplitude noise
        freq_p:      numpy array, interpreted freq_vector input
        pn_p:        numpy array, interpreted pn_vector input
    '''
    # check input
    if (freq_vector.shape != pn_vector.shape) or (freq_vector.shape[0] < 2) or \
       (fs <= 0) or (N < 1):
        return False, None, None, None

    # prepare the frequency vector of the data series
    df_bin      = fs / N                                # frequency bin size, Hz
    freq        = np.arange(N) * df_bin                 # frequency vector, Hz
    freq[freq > fs/2] -= fs                             # full freq vector with negative freq, Hz
    
    # interpolate the PSD at positive frequencies
    freq_p      = freq[freq > 0]                        # +freq (excl. DC but with fs/2 if exists), Hz
    pn_p        = np.interp(10*np.log10(freq_p),        # linear interpolation in log scale
                            10*np.log10(freq_vector),
                            pn_vector)
    
    # calculate the spectrum (see eq (6.15) of LLRF book) and add random phases
    pn_p_cplx   = np.sqrt(10**(pn_p/10) * N * fs / 2) * \
                  np.exp(1j * rand_unif(low = -np.pi, high = np.pi, n = pn_p.shape[0]))
    
    # get the full complex spectrum (0 to fs) of the phase noise
    if np.mod(N, 2) == 0:
        pn_spec = np.concatenate([np.array([pn_p_cplx[0]]),         # fs/2 bin exists
                                  pn_p_cplx,                              
                                  np.conj(pn_p_cplx[-2::-1])])
    else:
        pn_spec = np.concatenate([np.array([pn_p_cplx[0]]),         # fs/2 bin does not exist
                                  pn_p_cplx,                              
                                  np.conj(pn_p_cplx[::-1])])

    # calculate the noise time series        
    pn_series = np.real(np.fft.ifft(pn_spec))
    pn_series -= np.mean(pn_series)
    
    return True, pn_series, freq_p, pn_p

def calc_rms_from_psd(freq_vector, pn_vector, freq_start, freq_end, fs, N):
    '''
    calculate the rms value from PSDs.

    Parameters:
        freq_vector: numpy array, offset frequency from carrier, Hz
        pn_vector:   numpy array, DSB noise PSD, dBrad^2/Hz for phase noise
        freq_start:  float, starting frequency for integration, Hz
        freq_end:    float, ending frequency for integration, Hz
        fs:          float, sampling frequency, Hz
        N:           int, number of points in the freq integration range
        
    Returns:
        status:      boolean, success (True) or fail (False)
        freq_p:      numpy array, interpreted freq_vector input
        pn_p:        numpy array, interpreted pn_vector input
        jitter_p:    numpy array, integrated jitter starting from freq_start
                      to different freq_p element values
    '''
    # check input
    if (freq_vector.shape != pn_vector.shape) or (freq_vector.shape[0] < 2) or \
       (fs <= 0) or (N <= 0) or (freq_start <= 0) or (freq_end <= freq_start):
        return False, None, None, None
            
    # get the frequeny vector in the integration range and interpolate the PSD
    df_bin  = (freq_end - freq_start) / N
    freq_p  = np.linspace(freq_start, freq_end, N + 1)
    pn_p    = np.interp(10*np.log10(freq_p),            # linear interpolation in log scale
                        10*np.log10(freq_vector),
                        pn_vector)
    
    # integrate the jitter in the frequency integration region
    bin_dB      = 10.0 * np.log10(df_bin)
    jitter_p    = np.zeros(freq_p.shape)    
    noise_power = 0.0
    
    for i in range(N+1):
        jitter_p[i] = np.sqrt(noise_power)
        noise_power += 10.0**((pn_p[i] + bin_dB) / 10.0)    
    
    return True, freq_p, pn_p, jitter_p
        
def notch_filt(wf, fnotch, Q, fs, b = None, a = None):
    '''
    Apply notch filter to the signal.

    Parameters:
        wf:      numpy array, the waveform to be notch filtered
        fnotch:  float, frequency to be notched, Hz
        Q:       float, quality factor of the notch filter
        fs:      float, sampling frequency, Hz
        b, a:    numpy arrays filter coefficients. Note that the user can
                  either input ``b, a``, or ``fnotch, Q, fs``
                  
    Returns:
        status:  boolean, success (True) or fail (False)
        wf_f:    numpy array, filtered waveform
        b, a:    numpy array, filter coefficients. If the same filter 
                  is used to filter multiple waveforms, we can use it
                  to avoid repeat the filter synthesis
    '''
    # check the input
    if (wf is None) or (wf.shape[0] < 3) or (fnotch <= 0.0) or \
       (Q <= 0.0) or (fs <= 0.0):
        return False, None, None, None

    # create the notch filter if not available
    if (b is None) or (a is None):
        b, a = signal.iirnotch(fnotch, Q, fs)

    # filter the data
    return True, signal.filtfilt(b, a, wf), b, a

def design_notch_filter(fnotch, Q, fs):
    '''
    Design the notch filter (return a discrete filter).

    Parameters:
        fnotch:  float, frequency to be notched, Hz
        Q:       float, quality factor of the notch filter
        fs:      float, sampling frequency, Hz
        
    Returns:
        status:         boolean, success (True) or fail (False)
        Ad, Bd, Cd, Dd: numpy matrix, discrete notch filter
    '''
    # check the input
    if (fnotch <= 0.0) or (Q <= 0.0) or (fs <= 0.0):
        return (False,) + (None,)*4    

    # create the notch filter
    b, a = signal.iirnotch(fnotch, Q, fs)
    dsys = signal.dlti(b, a, dt = 1.0/fs)
    dsys = signal.StateSpace(dsys)
    Ad, Bd, Cd, Dd, dt = dsys.A, dsys.B, dsys.C, dsys.D, dsys.dt    

    # filter the data
    return True, Ad, Bd, Cd, Dd

def filt_step(Afd, Bfd, Cfd, Dfd, in_step, state_f0):
    '''
    Apply a step of the filter.

    Parameters:
        Afd, Bfd, Cfd, Dfd: numpy matrix, discrete filter model
        in_step:            float or complex, input of the time step
        state_f0:           numpy matrix, state of the filter of last step
        
    Returns:
        status:   boolean, success (True) or fail (False)
        out_step: float or complex, filter output of this time step
        state_f:  numpy matrix, state of the filter of this step, 
                   should input to the next execution    
    '''
    # calculate the controller output
    state_f  = Afd * state_f0 + Bfd * in_step
    out_step = Cfd * state_f0 + Dfd * in_step

    # return the results of the step
    return True, out_step, state_f

def moving_avg_obs(wf_in, n):
    '''
    Moving average without compensating the group delay (no longer used).
    
    Parameters:
        wf_in:   numpy array, input waveform
        n:       int, point number of moving average
        
    Returns:
        status:  boolean, success (True) or fail (False)
        wf_out:  numpy array, output waveform
    '''
    # check the input
    if n <= 0 or wf_in.shape[0] < 3:
        return False, None

    # make the moving average
    wf_out     = np.cumsum(wf_in, dtype = float)
    wf_out[n:] = (wf_out[n:] - wf_out[:-n]) / n
    wf_out[:n] = wf_in[:n]

    return True, wf_out
    
def rand_sine(N, fs, nfreq = 1, Amin = 0.0, Amax = 1.0, fmin = 0.0, fmax = 1e3):
    '''
    Generate a data series of the sum of several random sine signals.
    
    Parameters:
        N:          int, number of point of the series
        fs:         float, sampling frequency, Hz
        nfreq:      int, number of frequencies
        Amin, Amax: float, min and max values of amplitude
        fmin, fmax: float, min and max values of frequencies, Hz
        
    Returns:
        status:     boolean, success (True) or fail (False)
        sout:       numpy array, output data series
        t:          numpy array, time series, s
    '''
    # check the inputs
    if N <= 0 or fs <= 0 or nfreq < 1 or Amin < 0 or \
       Amax < Amin or fmin < 0 or fmax < fmin:
        return False, None, None, None
        
    # generate the random amplitude, phase and frequency
    A = np.random.uniform(Amin, Amax, nfreq)
    f = np.random.uniform(fmin, fmax, nfreq)
    P = np.random.uniform(-np.pi, np.pi, nfreq)

    # generate the series
    sout = np.zeros(N)
    t    = np.arange(N) / fs
    for i in range(len(A)):
        sout = sout + A[i] * np.sin(2.0 * np.pi * f[i] * t + P[i])

    # return
    return True, sout, t, (A, f, P)

def gen_rand_sine_from_psd(freq_vector, pn_vector, freqs):
    '''
    generate random sine functions from DSB PSD.

    Parameters:
        freq_vector: numpy array, offset frequency from carrier, Hz
        pn_vector:   numpy array, DSB noise PSD, dBrad^2/Hz for phase noise
        freqs:       numpy array, frequencies of sine waves (prepared by user), Hz
        
    Returns:
        status:      boolean, success (True) or fail (False)
        amplts:      numpy array, amplitudes of the sine waves
        phases:      numpy array, phases of the sine waves, rad
        psds:        numpy array, interpreted phase/amplitude noise DSB PSD at 
                                  selected frequencies, dBrad^2/Hz for phase noise
    '''
    # check input
    if (freq_vector.shape != pn_vector.shape) or (freq_vector.shape[0] < 2) or \
       (freqs.shape[0] < 1):
        return (False,) + (None,)*3

    # calculates the sine wave parameters
    n_sine = len(freqs)   
    psds   = np.interp(10 * np.log10(freqs), 
                       10 * np.log10(freq_vector), 
                       pn_vector)                               # PSDs at freqs, dBrad^2/Hz for phase noise
    bw_t   = (freqs[1:] + freqs[:-1]) / 2
    bw_t   = np.hstack(([0], bw_t, freqs[-1]))
    bws    = bw_t[1:] - bw_t[:-1]                               # bandwidth of each frequency points, Hz
    amplts = 10**((psds + 10*np.log10(bws) + 3) / 20)           # amplitudes of sine waves, rad
                                                                # reasons why we need this 3dB is not clear yet
    phases = np.random.rand(n_sine) * 2.0 * np.pi               # random phases of sine waves, rad
    
    # return the results
    return True, amplts, phases, psds

def moving_avg(data, n):
    '''
    moving average with group delay compensated.

    Parameters:
        data:   numpy array, input 1D waveform to be filtered
        n:      numpy array, number of points for average
        
    Returns:
        status:      boolean, success (True) or fail (False)
        result:      numpy array, filtered data
    '''
    # check the input
    if (len(data) <= n) or (n <= 1):
        return False, None

    # moving average
    result = uniform_filter1d(data, size = n)
    return True, result














"""Fit data to sine/cosine, circle, ellipse or Gaussian functions."""
#############################################################################
#  Copyright (c) 2023 by Paul Scherrer Institute, Switzerland
#  All rights reserved.
#  Authors: Zheqiao Geng
#############################################################################
'''
#########################################################################
Here collects routines for fitting algorithms

Implemented:
    - fit_sincos    : fit the sine or cosine function
    - fit_circle    : fit the 2D points to a circle function
    - fit_ellipse   : fit the 2D points to an ellipse function
    - fit_Gaussian  : fit a 1D Guassian function
#########################################################################
'''
import numpy as np
from scipy.optimize import curve_fit

def fit_sincos(X_rad, Y, target = 'cos'):
    '''
    Fit the sine or cosine function
      - ``y = A * cos(x + phi) + c = (A*cos(phi)) * cos(x) - (A*sin(phi)) * sin(x) + c``
      - ``y = A * sin(x + phi) + c = (A*cos(phi)) * sin(x) + (A*sin(phi)) * cos(x) + c``

    Parameters:
        X_rad:   numpy array, phase array in radian
        Y:       numpy array, value of the sine or cosine function
        target:  'sin' or 'cos', determine which function to fit to
        
    Returns:
        status:  boolean, success (True) or fail (False)
        A:       float, amplitude of the function
        phi_rad: float, phase of the function, rad
        c:       float, offset of the function
    '''
    # check the input
    if (not X_rad.shape == Y.shape) or (X_rad.shape[0] < 3):
        return False, None, None, None

    # make the least-square fitting
    if target == 'cos':
        P = np.linalg.lstsq(np.vstack((np.cos(X_rad), -np.sin(X_rad), np.ones(Y.shape))).T, Y, rcond = None)
    else:
        P = np.linalg.lstsq(np.vstack((np.sin(X_rad),  np.cos(X_rad), np.ones(Y.shape))).T, Y, rcond = None)

    # calculate the result
    A       = np.sqrt(P[0][0]**2 + P[0][1]**2)    # P[0][0] = A*cos(phi); P[0][1] = A*sin(phi)
    phi_rad = np.arctan2(P[0][1], P[0][0])
    c       = P[0][2]

    return True, A, phi_rad, c

def fit_circle(X, Y):
    '''
    Fit a circle. Given ``x``, ``y`` data points and find ``x0``, ``y0`` and ``r`` following
    ``(x - x0)^2 + (y - y0)^2 = r^2``.

    Parameters:
        X:      numpy array, x-coordinate of the points
        Y:      numpy array, y-coordinate of the points
        
    Returns:
        status: boolean, success (True) or fail (False)
        x0, y0: float, center coordinate of the circle
        r:      float, radius of the circle
    '''
    # check the input
    if (not X.shape == Y.shape) or (X.shape[0] < 3):
        return False, None, None, None

    # make the fit again with least square
    P = np.linalg.lstsq(np.vstack((2*X, 2*Y, np.ones(Y.shape))).T, X**2+Y**2, rcond = None)

    # calculate the result
    x0 = P[0][0]
    y0 = P[0][1]
    r  = np.sqrt(P[0][2] + x0**2 + y0**2)

    return True, x0, y0, r

def fit_ellipse(X, Y):
    '''
    Fit an ellipse. Get the general ellipse function and its characteristics, including
    semi-major axis ``a``, semi-minor axis ``b``, center coordinates ``(x0, y0)`` and rotation
    angle ``sita`` (the angle from the positive horizontal axis to the ellipse's major axis).
    The general ellipse equation is ``A*X^2 + B*X*Y + C*Y^2 + D*X + E*Y + F = 0``.
    When making the fitting, we divide the equation by ``C`` to normalize the coefficient of 
    ``Y^2`` to 1 and move it to the right side of the fitting equation. The ellipse is derived 
    with the following steps:
    
     1. define a canonical ellipse: 
            - ``X1 = a * cos(phi)``
            - ``Y1 = b * sin(phi)`` 
            
        where ``phi`` is a phase vector covering from 0 to 2*pi
        
     2. rotate the canonical ellipse: 
            - ``X2 = X1 * cos(sita) - Y1 * sin(sita)``
            - ``Y2 = X1 * sin(sita) + Y1 * cos(sita)``
            
     3. add offset to the rotated ellipse:
            - ``X = X2 + x0``
            - ``Y = Y2 + y0``

    See the webpage: https://en.wikipedia.org/wiki/Ellipse.  

    Parameters:
        X:      numpy array, x-coordinate of the points
        Y:      numpy array, y-coordinate of the points
        
    Returns:
        status: boolean, success (True) or fail (False)
        Coef:   list, coefficiets derived from the least-square fitting
        a:      float, semi-major
        b:      float, semi-minor
        x0, y0: float, center of the ellipse
        sita:   float, angle of the ellipse (see above), rad
    '''
    # check the input
    if (not X.shape == Y.shape) or (X.shape[0] < 3):
        return (False,) + (None,)*6

    # make the fit with least square
    P = np.linalg.lstsq(np.vstack((X**2, X*Y, X, Y, np.ones(Y.shape))).T, -Y**2, rcond = None)

    # calculate the characteristics
    A = P[0][0]
    B = P[0][1]
    C = 1.0
    D = P[0][2]
    E = P[0][3]
    F = P[0][4]

    a  = -np.sqrt(2*(A*E**2 + C*D**2 - B*D*E + (B**2 - 4*A*C)*F)*((A+C) + np.sqrt((A-C)**2 + B**2))) / (B**2 - 4*A*C)
    b  = -np.sqrt(2*(A*E**2 + C*D**2 - B*D*E + (B**2 - 4*A*C)*F)*((A+C) - np.sqrt((A-C)**2 + B**2))) / (B**2 - 4*A*C)
    x0 = (2*C*D - B*E) / (B**2 - 4*A*C)
    y0 = (2*A*E - B*D) / (B**2 - 4*A*C)

    if B != 0.0:
        sita = np.arctan((C - A - np.sqrt((A-C)**2 + B**2)) / B)
    else:
        if A <= C:
            sita = 0
        else:
            sita = np.pi/2

    return True, [A,B,C,D,E,F], a, b, x0, y0, sita

def fit_Gaussian(X, Y):
    '''
    Fit a Guassian distribution function. 
    
    See the webpage: https://pythonguides.com/scipy-normal-distribution/.

    Parameters:
        X:      numpy array, x-coordinate of the points
        Y:      numpy array, y-coordinate of the points
        
    Returns:
        status: boolean, success (True) or fail (False)
        a:      float, magnitude scale of the un-normalized distribution
        mu:     float, mean value
        sigma:  float, standard deviation
    '''
    # check the input
    if (not X.shape == Y.shape) or (X.shape[0] < 3):
        return (False,) + (None,)*3
    
    # define the Gaussian (normal) distribution 
    def norm_dist(x, a, mu, var):
        y = a * np.exp(-0.5 / var * (x - mu)**2)    # since X,Y are not normalized, "a" can be arbitrary value
        return y

    # get the initial guess of the parameters
    mu_est  = sum(X * Y) / sum(Y)
    var_est = sum((X - mu_est)**2 * Y) / sum(Y)

    # make the fit
    P, cov  = curve_fit(norm_dist, X, Y, p0 = [1.0, mu_est, var_est])
    a, mu, var = P[0], P[1], P[2]

    return True, a, mu, np.sqrt(var)






















"""Plotting functions for internal use."""
#############################################################################
#  Copyright (c) 2023 by Paul Scherrer Institute, Switzerland
#  All rights reserved.
#  Authors: Zheqiao Geng
#############################################################################
'''
#########################################################################
Here collects all plotting functions. The goal is to avoid importing the 
"matplotlib" module in the algorithm codes to enable running the algorithms
in an embedded CPU where the "matplotlib" is not installed
#########################################################################
'''
import numpy as np
import matplotlib.pyplot as plt

def plot_ss_discrete(fc, Ac_dB, Pc_deg, fd, Ad_dB, Pd_deg, Ts):
    '''
    Plot the frequency responses of the continous and discrete state-space
    equations, used in function ``ss_discrete`` of the ``rf_control`` module.
    '''
    plt.figure()
    plt.subplot(2,1,1)
    plt.plot(fc, Ac_dB, label = 'Continous')
    plt.plot(fd, Ad_dB, label = 'Discrete')
    plt.axvline( 0.5 / Ts, ls = '--')
    plt.axvline(-0.5 / Ts, ls = '--')
    plt.legend()
    plt.grid()
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Magnitude (dB)')
    plt.subplot(2,1,2)
    plt.plot(fc, Pc_deg, label = 'Continous')
    plt.plot(fd, Pd_deg, label = 'Discrete')
    plt.axvline( 0.5 / Ts, ls = '--')
    plt.axvline(-0.5 / Ts, ls = '--')
    plt.legend()
    plt.grid()
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Phase (deg)')
    plt.suptitle('State-space System Discretization')
    plt.show(block = False)
    
def plot_ss_freqresp(f_wf, A_wf_dB, P_wf_deg, h, fs, Ts, title):
    '''
    Plot the frequency response of a state-space system, used in function 
    ``ss_freqresp`` of the ``rf_control`` module.
    '''
    plt.figure()
    plt.subplot(2,2,1)                          # bode plot
    plt.plot(f_wf, A_wf_dB)
    if Ts is not None:
        plt.axvline( fs / 2, ls = '--')
        plt.axvline(-fs / 2, ls = '--')
    plt.grid()
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Magnitude (dB)')
    plt.subplot(2,2,3)
    plt.plot(f_wf, P_wf_deg)
    if Ts is not None:
        plt.axvline( fs / 2, ls = '--')
        plt.axvline(-fs / 2, ls = '--')
    plt.grid()
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Phase (deg)')
    plt.subplot(1,2,2)                          # Nyquist plot (need to plot +/- frequencies differently)
    plt.plot(np.real(h[f_wf >= 0.0]), np.imag(h[f_wf >= 0.0]), label = 'Positive Frequency')
    plt.plot(np.real(h[f_wf <  0.0]), np.imag(h[f_wf <  0.0]), label = 'Negative Frequency')
    plt.plot([-1], [0], '*')
    plt.legend()
    plt.grid()
    plt.xlabel('Real')
    plt.ylabel('Imag')
    plt.suptitle(title)
    plt.show(block = False)
    
def plot_basic_rf_controller(w, h):
    '''
    Plot the frequency response of the basic RF controller, used in function 
    ``basic_rf_controller`` of the ``rf_control`` module.
    '''
    plt.figure()
    plt.subplot(2,1,1)
    plt.plot(w / 2 / np.pi, 20*np.log10(np.abs(h)), label = 'Controller')
    plt.legend()
    plt.grid()
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Magnitude (dB)')
    plt.subplot(2,1,2)
    plt.plot(w / 2 / np.pi, np.angle(h, deg = True), label = 'Controller')
    plt.legend()
    plt.grid()
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Phase (deg)')
    plt.suptitle('Basic RF Controller')
    plt.show(block = False)

def plot_loop_analysis(f_wf, L, S, T, fs, Ts, label):
    '''
    Plot the loop analysis results, used in function ``loop_analysis`` of the 
    ``rf_control`` module.
    '''
    plt.figure()
    plt.subplot(2,2,1)                          # bode plot
    plt.plot(f_wf, 20*np.log10(np.abs(L)),       label = 'L')
    plt.plot(f_wf, 20*np.log10(np.abs(S)), '-.', label = 'S')
    plt.plot(f_wf, 20*np.log10(np.abs(T)), ':',  label = 'T')
    if Ts is not None:
        plt.axvline( fs / 2, ls = '--')
        plt.axvline(-fs / 2, ls = '--')
    plt.legend()
    plt.grid()
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Magnitude (dB)')
    plt.title('Amplitude Response of L/S/T')

    plt.subplot(2,2,3)
    plt.plot(f_wf, np.angle(L, deg = True),       label = 'L')
    plt.plot(f_wf, np.angle(S, deg = True), '-.', label = 'S')
    plt.plot(f_wf, np.angle(T, deg = True), ':',  label = 'T')
    if Ts is not None:
        plt.axvline( fs / 2, ls = '--')
        plt.axvline(-fs / 2, ls = '--')
    plt.axhline( 180, ls = '--')
    plt.axhline(-180, ls = '--')
    plt.legend()
    plt.grid()
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Phase (deg)')
    plt.title('Phase Response of L/S/T')

    plt.subplot(1,2,2)                          # Nyquist plot (need to plot +/- frequencies differently)
    plt.plot(np.real(L[f_wf >= 0.0]), np.imag(L[f_wf >= 0.0]), label = 'Positive Frequency')
    plt.plot(np.real(L[f_wf <  0.0]), np.imag(L[f_wf <  0.0]), label = 'Negative Frequency')
    plt.plot([-1], [0], '*')
    plt.legend()
    plt.grid()
    plt.xlabel('Real')
    plt.ylabel('Imag')
    plt.title('Nyquist Plot of L')

    plt.suptitle('Feedback Loop Analysis ' + label)
    plt.show(block = False)
    
def plot_calc_psd(result):
    '''
    Plot the spectrum, used in function ``calc_psd_coherent`` and ``calc_psd`` of the 
    ``rf_noise`` module.
    '''
    plt.figure()
    plt.plot(result['freq'], result['amp_resp'])
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Magnitude (dBFS/Hz)')
    plt.grid()
    plt.show(block = False)        

def plot_cav_ss(wrf, hrf, wbm, hbm):
    '''
    Plot the spectra of cavity model, used in function ``cav_ss`` of the 
    ``rf_sim`` module.
    '''
    plt.figure()
    plt.subplot(2,1,1)
    plt.plot(wrf / 2 / np.pi, 20*np.log10(np.abs(hrf)), label = 'Cavity resp. to RF drive')
    plt.plot(wbm / 2 / np.pi, 20*np.log10(np.abs(hbm)), label = 'Cavity resp. to beam drive')
    plt.legend()
    plt.grid()
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Magnitude (dB)')
    plt.subplot(2,1,2)
    plt.plot(wrf / 2 / np.pi, np.angle(hrf, deg = True), label = 'Cavity resp. to RF drive')
    plt.plot(wbm / 2 / np.pi, np.angle(hbm, deg = True), label = 'Cavity resp. to beam drive')
    plt.legend()
    plt.grid()
    plt.xlabel('Frequency (Hz)')
    plt.ylabel('Phase (deg)')
    plt.suptitle('Cavity State-space Model')
    plt.show(block = False)

def plot_rf_power_req(Pfor, Pref, QL_vec):
    '''
    Plot the required RF power, used in function ``rf_power_req`` of the 
    ``rf_sim`` module.
    '''
    plt.figure()
    plt.subplot(1,2,1)
    for dw in Pfor.keys():
        plt.plot(QL_vec, Pfor[dw] / 1000.0, 
                 label = 'Detuning = {:.1f} Hz'.format(dw/2/np.pi))
    plt.legend()
    plt.grid()
    plt.xlabel(r'$Q_L$')
    plt.ylabel(r'$P_{for}$ (kW)')
    plt.subplot(1,2,2)
    for dw in Pfor.keys():
        plt.plot(QL_vec, Pref[dw] / 1000.0, 
                 label = 'Detuning = {:.1f} Hz'.format(dw/2/np.pi))
    plt.legend()
    plt.grid()
    plt.xlabel(r'$Q_L$')
    plt.ylabel(r'$P_{ref}$ (kW)')
    plt.show(block = False)

def plot_plot_ellipse(X, Y):
    '''
    Plot the ellipse, used in function ``plot_ellipse`` of the ``rf_misc`` module.
    '''
    plt.figure()
    plt.plot(X, Y, '-*')
    plt.xlabel('X')
    plt.ylabel('Y')
    plt.grid()
    plt.axis('equal')
    plt.show(block = False)

def plot_plot_Guassian(X, Y):
    '''
    Plot the Guassian distribution, used in function ``plot_Guassian`` of the 
    ``rf_misc`` module.
    '''
    plt.figure()
    plt.plot(X, Y, '-*')
    plt.xlabel('X')
    plt.ylabel('Probability (not normalized)')
    plt.grid()
    plt.show(block = False)












    


"""Design and analyze RF feedback/feedforward controllers."""
#############################################################################
#  Copyright (c) 2023 by Paul Scherrer Institute, Switzerland
#  All rights reserved.
#  Authors: Zheqiao Geng
#############################################################################
'''
#########################################################################
Here collects routines for RF controller design, analysis and simulation

Implemented:
    - ss_discrete         : discretize a continous state-space system and compare freq responses
    - ss_cascade          : cascade two state-space systems (either continous or discrete - C/D)
    - ss_freqresp         : calculate and plot freq response of a state-space system (C/D)
    - basic_rf_controller : derive a basic continous RF I/Q controller: P + I + frequency notches
    - control_step        : perform one time-step execution of the discretized controller
    - loop_analysis       : analyze the sensitivity/complementary sensitivity of an RF control loop (C/D)
    - cav_sp_ff           : derive the setpoint and feedforward waveforms for desired cavity voltage
                            and beam loading
    - ADRC_controller     : derive a basic ADRC controller (the observer and gain)
    - ADRC_control_step   : perform one time-step execution of the discretized controller including
                            the ADRC observer
    - AFF_timerev_lpf     : time-reversed low pass filter-based adaptive feedforward
    - AFF_ilc_design      : derive the ILC gain matrix from the impulse response and weighting
    - AFF_ilc             : apply the ILC algorithm to calculate the feedforward correction signal
    - resp_inv_svd        : response matrix inversion with SVD (with singular value filtering)
    - resp_inv_lsm        : response matrix inversion with lease-square method (with regularization)

To be implemented:
    - in "cav_sp_ff", smooth the feedforward and update the setpoint correspondingly
    - pulsed piezo drive generation
    - Active noise cancellation for tuning control (narro-band/wideband ANC)
    - look-up table based operating point determination
 
Some algorithms are referred to the following books:
S. Simrock and Z. Geng, Low-Level Radio Frequency Systems, Springer, 2022
https://link.springer.com/book/10.1007/978-3-030-94419-3 ("LLRF Book")
Z. Geng and S. Simrock, Intelligent Beam Control in Accelerators, Springer, 2023
https://link.springer.com/book/10.1007/978-3-031-28597-4 ("Beam Control Book")
#########################################################################
'''
import numpy as np
from scipy import signal
from numpy.linalg import matrix_rank

from llrflibs.rf_sysid import *
from llrflibs.rf_misc import *

def ss_discrete(Ac, Bc, Cc, Dc, Ts, method = 'zoh', alpha = 0.3, plot = False, plot_pno = 1000, spec_data = False):
    '''
    Derive the discrete state-space equation from a continous one and compare 
    their frequency responses.
    
    Parameters:
        Ac, Bc, Cc, Dc: numpy matrix (complex), continous state-space model
        Ts:             float, sampling time, s
        method:         string, ``gbt``, ``bilinear``, ``euler``, ``backward_diff``, ``zoh``, 
                         ``foh`` or ``impulse`` (see document of signal.cont2discrete)
        alpha:          float, 0 to 1 (see document of signal.cont2discrete) 
        plot:           boolean, enable the plot of frequency responses
        plot_pno:       int, number of point in the plot
        spec_data:      boolean, return spectra data or not
        
    Returns:
        status:         boolean, success (True) or fail (False)
        Ad, Bd, Cd, Dd: numpy matrix (complex), discrete state-space model
    '''
    # check the input
    if (Ts <= 0):
        return (False,) + (None,)*5

    plot_pno = 1000 if (plot_pno <= 0) else plot_pno

    # discretize it
    Ad, Bd, Cd, Dd, _ = signal.cont2discrete((Ac, Bc, Cc, Dc), Ts, method = method, alpha = alpha)

    # calculate the responses of both continous and discrete versions
    if plot or spec_data:
        sc, fc, Ac_dB, Pc_deg, _ = ss_freqresp(Ac, Bc, Cc, Dc, plot_pno = plot_pno, plot_maxf = 1.0 / Ts)
        sd, fd, Ad_dB, Pd_deg, _ = ss_freqresp(Ad, Bd, Cd, Dd, plot_pno = plot_pno, Ts = Ts)

        spec = {'fc': fc, 'Ac_dB': Ac_dB, 'Pc_deg': Pc_deg,
                'fd': fd, 'Ad_dB': Ad_dB, 'Pd_deg': Pd_deg}
    else:
        spec = None

    # plot the responses
    if plot:
        # make the plot
        if sc and sd:
            from llrflibs.rf_plot import plot_ss_discrete
            plot_ss_discrete(fc, Ac_dB, Pc_deg, fd, Ad_dB, Pd_deg, Ts)

    # return the results
    return True, Ad, Bd, Cd, Dd, spec

def ss_cascade(A1, B1, C1, D1, A2, B2, C2, D2, Ts = None):
    '''
    Cascade two state-space systems with the first system applyed to
    the input first. This function works for both continous system (``Ts`` is None)
    and discrete systems (``Ts`` has a nonzero floating value).
    
    Parameters:
        A1, B1, C1, D1: numpy matrix (complex), state-space model of system 1
        A2, B2, C2, D2: numpy matrix (complex), state-space model of system 2
        Ts:             float, sampling time, s
        
    Returns:
        status          boolean, success (True) or fail (False)
        A, B, C, D:     numpy matrix (complex), state-space model of cascaded system
    '''
    # check the input
    if Ts is not None:
        if Ts <= 0.0:
            return (False,) + (None,)*5

    # construct the state-space object
    if Ts is None:                      # continuous
        sys1 = signal.StateSpace(A1, B1, C1, D1)
        sys2 = signal.StateSpace(A2, B2, C2, D2)
    else:                               # discrete
        sys1 = signal.StateSpace(A1, B1, C1, D1, dt = Ts)
        sys2 = signal.StateSpace(A2, B2, C2, D2, dt = Ts)

    # cascade the system
    sys  = sys2 * sys1

    # return the results
    return True, sys.A, sys.B, sys.C, sys.D, Ts

def ss_freqresp(A, B, C, D, Ts = None, plot = False, plot_pno = 1000, plot_maxf = 0.0, 
                title = 'Frequency Response'):
    '''
    Plot the frequency response of a state-space system. This function works 
    for both continous system (``Ts`` is None) and discrete systems (``Ts`` has a 
    nonzero floating value).
    
    Parameters:
        A, B, C, D: numpy matrix (complex), state-space model of system
        Ts:         float, sampling time, s
        plot:       boolean, enable the plot of frequency response
        plot_pno:   int, number of point in the plot
        plot_maxf:  float, frequency range (+-) to be plotted, Hz
        title:      string, title showed on the plot
        
    Returns:
        status:     boolean, success (True) or fail (False)
        f_wf:       numpy array, frequency waveform, Hz
        A_wf_dB:    numpy array, amplitude response waveform, dB
        P_wf_deg:   numpy array, phase response waveform, deg
        h:          numpy array (complex), complex response
    '''    
    # check the input
    plot_pno  = 1000 if (plot_pno  <= 0) else plot_pno
    plot_maxf = 1e6  if (plot_maxf <= 0) else plot_maxf

    if Ts is not None:
        if Ts <= 0.0:
            return (False,) + (None,)*4

    # calculate the frequency response 
    if Ts is None:                      # continous
        maxw = 2 * np.pi * plot_maxf
        fs   = 1.0
        w, h = signal.freqresp((A, B, C, D), w = np.linspace(-maxw, maxw, plot_pno))
    else:                               # discrete
        maxw = np.pi
        fs   = 1.0 / Ts
        w, h = signal.dfreqresp((A, B, C, D, Ts), w = np.linspace(-maxw, maxw, plot_pno))

    # calculate the results for display
    f_wf     = w / 2 / np.pi * fs
    A_wf_dB  = 20 * np.log10(np.abs(h))
    P_wf_deg = np.angle(h, deg = True)

    # plot 
    if plot:
        from llrflibs.rf_plot import plot_ss_freqresp
        plot_ss_freqresp(f_wf, A_wf_dB, P_wf_deg, h, fs, Ts, title)

    # return the frequency response (frequency in absolute Hz)
    return True, f_wf, A_wf_dB, P_wf_deg, h

def basic_rf_controller(Kp, Ki, notch_conf = None, plot = False, plot_pno = 1000, plot_maxf = 0.0):
    '''
    Generate the continous state-space equation for a basic controller with
     * I/Q control strategy (input/output are all complex signals).
     * configurable for PI control + frequency notches.
    
    Parameters:
        Kp:          float, proportional (P) gain
        Ki:          float, integral (I) gain
        notch_conf:  dict, with the following items:
            ``freq_offs``: list, offset frequency to be notched, Hz; 
            ``gain``: list, gain of the notch (inverse of suppress ratio); 
            ``half_bw``: list, half bandwidth of the notch filter, rad/s
        plot:        boolean, enable the plot of frequency response of the controller
        plot_pno:    int, number of point in the plot
        plot_maxf:   float, frequency range (+-) to be plotted, Hz   
        
    Returns:
        status:             boolean, success (True) or fail (False)
        Akc, Bkc, Ckc, Dkc: numpy matrix (complex), continuous state-space controller
    '''
    # check the input
    if (Kp < 0) or (Ki < 0):
        return (False,) + (None,)*4

    if notch_conf is not None:
        if (not isinstance(notch_conf, dict)) or \
           ('freq_offs' not in notch_conf.keys()) or \
           ('gain'      not in notch_conf.keys()) or \
           ('half_bw'   not in notch_conf.keys()): 
            return (False,) + (None,)*4

    plot_pno  = 1000 if (plot_pno  <= 0) else plot_pno
    plot_maxf = 1e6  if (plot_maxf <= 0) else plot_maxf

    # first get the PI controller
    K_num, K_den = add_tf([Kp], [1.0], [Ki], [1.0, 0.0])

    # add the notch filters
    if notch_conf is not None:
        # get the notch parameters
        nt_f  = notch_conf['freq_offs']     # frequency offset, Hz
        nt_g  = notch_conf['gain']          # gain
        nt_wh = notch_conf['half_bw']       # half bandwidth of the notch filter, rad/s

        # add the notch filter transfer functions
        for i in range(len(nt_f)):
            K_num, K_den = add_tf(K_num, K_den, 
                                  [nt_g[i] * nt_wh[i]], 
                                  [1.0, nt_wh[i] - 1j*2*np.pi*nt_f[i]])
            K_num, K_den = add_tf(K_num, K_den, 
                                  [np.conj(nt_g[i]) * nt_wh[i]], 
                                  [1.0, nt_wh[i] + 1j*2*np.pi*nt_f[i]])

    # get the state-space model
    Akc, Bkc, Ckc, Dkc = signal.tf2ss(K_num, K_den)

    # plot the response
    if plot:
        # calculate the response
        w, h = signal.freqs(K_num, K_den, 
                            worN = np.linspace(-2*np.pi*plot_maxf, 
                                                2*np.pi*plot_maxf, 
                                                plot_pno))

        # make the plot
        from llrflibs.rf_plot import plot_basic_rf_controller
        plot_basic_rf_controller(w, h)

    # return the results
    return True, Akc, Bkc, Ckc, Dkc

def control_step(Akd, Bkd, Ckd, Dkd, err_step, state_k0, ff_step = 0.0):
    '''
    Controller execute for one step based on the discrete state-space equation.
    
    Parameters:
        Akd, Bkd, Ckd, Dkd: numpy matrix (complex), discrete state-space controller
        err_step:           complex, system output error (input to controller) of this step
        state_k0:           numpy matrix (complex), state of the last step
        ff_step:            complex, feedforward of this step
        
    Returns:
        status:             boolean, success (True) or fail (False)
        ctrl_step:          complex, overall output of the controller of this step
        ctrl_out:           complex, feedback output (exclude feedforward wrt ctrl_step)
        state_k:            numpy matrix (complex), state of this step (should be input to 
                             the next execution of this function)
        
    Question: 
        In the second equation, shall we use ``state_k0`` or ``state_k``?
    '''
    # calculate the controller output
    state_k   = Akd * state_k0 + Bkd * err_step
    ctrl_out  = Ckd * state_k0 + Dkd * err_step
    ctrl_step = ctrl_out + ff_step

    # return the results of the step
    return True, ctrl_step, ctrl_out, state_k

def loop_analysis(AG, BG, CG, DG, AK, BK, CK, DK, Ts = None, delay_s = 0, 
                  plot = True, plot_pno = 100000, plot_maxf = 0.0, label = ''):
    '''
    Control loop analysis, including
     * derive the open loop transfer function.
     * calculate the senstivity and complementary sensitivity functions. 
     
    This function works for both continous system (``Ts`` is None) and discrete 
    systems (``Ts`` has a nonzero floating value).
    
    Parameters:
        AG, BG, CG, DG: numpy matrix (complex), plant model
        AK, BK, CK, DK: numpy matrix (complex), controller
        Ts:             float, sampling time, s
        delay_s:        float, loop delay, s        
        plot:           boolean, enable the plot of bode and Nyquist plots
        plot_pno:       int, number of point in the plot
        plot_maxf:      float, frequency range (+-) to be plotted, Hz    
        
    Returns:
        status:         boolean, success (True) or fail (False)
        S_max:          float, maximum sensitivity, dB
        T_max:          float, maximum complementary sensitivity, dB
    '''
    # check the input
    plot_pno  = 100000 if (plot_pno  <= 0) else plot_pno
    plot_maxf = 1e6    if (plot_maxf <= 0) else plot_maxf
    delay_s   = 0.0    if (delay_s   <  0) else delay_s

    if Ts is not None:
        if Ts <= 0.0:
            return False, None, None

    # parameter based on continous or discrete (Ts not None)
    fs = 1.0 if (Ts is None) else 1.0/Ts

    # cascade the plant and the controller get the open loop L
    status, AL, BL, CL, DL, _ = ss_cascade(AK, BK, CK, DK, AG, BG, CG, DG, Ts = Ts)
    if not status:
        return False, None, None

    # frequency response of L
    result_L    = ss_freqresp(AL, BL, CL, DL, Ts = Ts, 
                              plot_pno = plot_pno, plot_maxf = plot_maxf)
    status      = result_L[0]
    f_wf        = result_L[1]
    AL_dB       = result_L[2]
    L           = result_L[4]                       # complex response of L

    if not status:
        return False, None, None

    # apply the delay to the frequency response of L
    L *= np.exp(-1j * 2.0 * np.pi * f_wf * delay_s)

    # sensitivity and complementary sensitivity
    S = 1.0 / (1.0 + L)
    T = 1.0 - S

    # plot (similar to ss_freqresp)
    if plot:
        from llrflibs.rf_plot import plot_loop_analysis
        plot_loop_analysis(f_wf, L, S, T, fs, Ts, label)

    # calculate the peak of S and T
    S_max = np.max(20*np.log10(np.abs(S)))
    T_max = np.max(20*np.log10(np.abs(T)))

    return True, S_max, T_max

def cav_sp_ff(half_bw, filling_len, flattop_len, Ts, pno,
                vc0        = 1.0,
                detuning   = 0.0, 
                beta       = 1e4, 
                const_fpow = True,
                ib0        = None,
                phib_deg   = 0.0,
                beam_ids   = 0,
                beam_ide   = 0,
                roQ_or_RoQ = 0.0,
                QL         = 3e6,
                machine    = 'linac'):
    '''
    Generate the basic cavity set point and feedforward used to configure
    the LLRF controller (later may apply smooth in the edges).
    
    Refer to LLRF Book section 9.2.1.
    
    Parameters:
        half_bw:     float, half bandwidth of the cavity, rad/s
        filling_len: int, length of cavity filling time, number of samples
        flattop_len: int, length of the flattop for beam acc., number of samples
        Ts:          float, sampling time, s
        pno:         int, number of samples in the returned waveforms
        vc0:         float, desired cavity voltage at the flattop, V
        detuning:    float, detuning of the cavity, rad/s
        beta:        float, input coupling factor (needed for NC cavities; 
                      for SC cavities, can use the default value, or you can 
                      specify it if more accurate result is needed)
        const_fpow:  boolean, True for forcing constant filling drive power/phase
        ib0:         float, average beam current, A
        phib_deg:    float, beam accelerating phase (0 for on-crest), deg
        beam_ids:    int, beam starting sample index
        beam_ide:    int, beam ending sample index
        roQ_or_RoQ:  r/Q of Linac or R/Q of circular accelerator, Ohm
                      (see the note below)
        QL:          float, loaded quality factor of the cavity
        machine:     string, ``linac`` or ``circular``, used to select r/Q or R/Q
        
    Returns:
        status:      boolean, success (True) or fail (False)
        sp:          numpy array (complex), set point waveform (for controller)
        vf_ff:       numpy array (complex), feedforward waveform (for controller)
        vb:          numpy array (complex), beam drive voltage waveform
        Tpul:        numpy array, time array for the waveforms
    
    Note: 
          Linacs define the ``r/Q = Vacc**2 / (w0 * U)`` while circular machines 
          define ``R/Q = Vacc**2 / (2 * w0 * U)``, where ``Vacc`` is the accelerating
          voltage, ``w0`` is the angular cavity resonance frequency and ``U`` is the
          cavity energy storage. Therefore, to use this function, one needs to 
          specify the ``machine`` to be ``linac`` or ``circular``. Generally, we have
          ``R/Q = 1/2 * r/Q``.
    '''
    # check the input
    if (half_bw <= 0) or (filling_len <= 0) or (flattop_len <= 0) or \
       (Ts <= 0) or (pno < filling_len + flattop_len):
        return (False,) + (None,)*4

    vc0  = 1.0 if (vc0  <= 0.0) else vc0
    beta = 1e4 if (beta <= 0.0) else beta
    
    if ib0 is not None:
        if (ib0 < 0.0) or (beam_ids < 0) or (beam_ide <= beam_ids) or \
           (roQ_or_RoQ <= 0.0) or (QL <= 0.0):
            return (False,) + (None,)*4

    # define the setpoint table
    N  = filling_len + flattop_len          # effective pulse width
    T  = np.arange(filling_len+1) * Ts
    sp = np.ones(N, dtype = complex) 
    sp[:filling_len+1] = (1.0 - np.exp(-half_bw * T)) / (1.0 - np.exp(-half_bw * T[-1]))
    sp *= vc0
    
    # derive the ff of vf
    status, vf_ff, _ = cav_drv_est(sp, half_bw, Ts, detuning, beta)
    if not status:
        return (False,) + (None,)*4

    # handle the filling power if require to be constant (see Example 3.1 of LLRF Book)
    if const_fpow:
        vf_fill = vc0 * (half_bw - 1j*detuning) / \
                  (1.0 - np.exp(-(half_bw - 1j*detuning) * T[-1])) * \
                  (beta + 1) / (2 * half_bw * beta)
        vf_ff[:filling_len+1] = vf_fill

    # calculate the beam drive voltage
    vb = np.zeros(vf_ff.shape, dtype = complex)
    if ib0 is not None:
        # calculate the loaded resistance
        if machine == 'circular':
            RL = roQ_or_RoQ * QL
        else:
            RL = 0.5 * roQ_or_RoQ * QL
       
        # determine the beam drive
        vb[beam_ids:beam_ide+1] = RL * ib0 * np.exp(1j*(np.pi - phib_deg*np.pi/180.0))

    # remove the beam term
    vf_ff -= (beta + 1) / beta * vb

    # attend the pulse to the desired length
    if pno > N:
        sp    = np.hstack((sp,    np.zeros(pno - N)))
        vf_ff = np.hstack((vf_ff, np.zeros(pno - N)))
        vb    = np.hstack((vb,    np.zeros(pno - N)))

    return True, sp, vf_ff, vb, np.arange(pno)*Ts

def ADRC_controller(half_bw, pole_scale = 50.0):
    '''
    Generate the continous ADRC controller. We assume that the system gain
    has been normalized to 1 and the system phase corrected to 0. Therefore,
    the referred cavity equation is
    ``dvc/dt + (half_bw - 1j * detuning)*vc = half_bw * vd``
    where ``vc`` is the vector of cavity voltage and ``vd`` is the vector of cavity 
    drive (in principle ``vd = 2 * beta * vf / (beta + 1)``).
    
    Refer to LLRF Book section 4.2.3.
    
    Parameters:
        half_bw:    float, half bandwidth of the cavity, rad/s
        pole_scale: float, define the pole location of the observer
        
    Returns:
        status:                 boolean, success (True) or fail (False)
        Aobc, Bobc, Cobc, Dobc: numpy matrix (complex), continous ADRC observer
        b0:                     float, ADRC gain
    '''
    # check the input
    if (half_bw <= 0) or (pole_scale <= 0):
        return (False,) + (None,)*5

    # parameters
    p_obs = -pole_scale * half_bw           # pole of the observer
    l1    = -2 * p_obs                      # observer matrix parameter
    l2    = p_obs**2                        # observer matrix parameter
    b0    = half_bw                         # gain for RF drive voltage

    # construct the continous ADRC observer
    Aobc = np.matrix([[-l1, 1],
                      [-l2, 0]], dtype = complex)
    Bobc = np.matrix([[ l1, b0],
                      [ l2,  0]], dtype = complex)
    Cobc = Dobc = np.matrix(np.zeros(Aobc.shape), dtype = complex)

    # return the ADRC observer and gain
    return True, Aobc, Bobc, Cobc, Dobc, b0

def ADRC_control_step(Akd, Bkd, Ckd, Dkd, 
                      Aobd, Bobd, b0, 
                      sp_step, vc_step, vd_step,
                      state_k0, state_ob0, 
                      vf_step = 0.0, 
                      ff_step = 0.0, 
                      apply_to_err = False):
    '''
    Controller execute for one step based on the controller's discrete state-space equation
    including the ADRC observer.
    
    Refer to LLRF Book Figure 4.6.
    
    Parameters:
        Akd, Bkd, Ckd, Dkd: numpy matrix (complex), discrete RF controller
        Aobd, Bobd:         numpy matrix (complex), discrete ADRC observer (C,D not needed)
        b0:                 float, ADRC gain
        sp_step:            complex, cavity voltage setpoint of this time step
        vc_step:            complex, cavity voltage meas. of last time step
        vd_step:            complex, cavity drive of last time step
        vf_step:            complex, feedfoward part of cavity drive of last time step
        state_k0:           numpy matrix (complex), controller state of last time step
        state_ob0:          numpy matrix (complex), ADRC observer state of last time step
        ff_step:            complex, feedforward of this time step
        apply_to_err:       boolean, True to apply ADRC to error, or apply to whole cavity voltage
        
    Returns:
        status:             boolean, success (True) or fail (False)
        ctrl_step:          complex, overall output of the controller of this step
        ctrl_out:           complex, feedback output (exclude feedforward wrt ctrl_step)
        state_k:            numpy matrix (complex), controller state of this step (should input to 
                             the next execution of this function)
        state_ob:           numpy matrix (complex), ADRC observer state of this step (should input
                             to the next execution of this function)
        f:                  complex, estimated general disturbance by the ADRC observer
        
    Question: 
        In the second equation, shall we use ``state_k0`` or ``state_k``?
    
    Note: 
          1. looks like ADRC can mitigate the instability caused by the cavity's
             passband modes, then no notch filter is required to filter the cavity 
             voltage measurement.
          2. basic ADRC is perfect with a proportional controller - resulting
             in zero steady-state error. Looks like with other controller (e.g.,
             PI control), ADRC needs to be revised for good performance.
          3. adding feedforward to the basic ADRC also causes some problem, like
             the steady state error becomes nonzero.
          4. The solution to handle the problems of point 2 and 3 above is to apply
             ADRC to the errors of the cavity drive and voltage. See the paper:
             S. Zhao et al, Tracking and disturbance rejection in non-minimum
             phase systems, Proceedings of the 33rd Chinese Control Conference, 
             pp. 3834-3839, July 28-30, 2014, Nanjing, China.
    '''
    # execute one step based on user preference
    if apply_to_err:
        # execute observer and estimate cavity voltage error (vc_err_est) and disturbance (f)
        state_ob = Aobd * state_ob0 + Bobd * np.matrix([[vc_step - sp_step], [vd_step - vf_step]])
        vc_err_est, f = -state_ob[0, 0], state_ob[1, 0]

        # calculate the controller output
        state_k  = Akd * state_k0 + Bkd * vc_err_est
        ctrl_out = Ckd * state_k0 + Dkd * vc_err_est

        # calculate the final drive to the cavity (slightly different from Fig.4.6, 
        # we do not divide controller output by b0)
        ctrl_step = ctrl_out - f / b0 + ff_step
        vc_est    = sp_step - vc_err_est

    else:
        # execute observer and estimate cavity voltage (vc_est) and disturbance (f)
        state_ob = Aobd * state_ob0 + Bobd * np.matrix([[vc_step], [vd_step]])
        vc_est, f = state_ob[0, 0], state_ob[1, 0]

        # calculate the controller output
        state_k  = Akd * state_k0 + Bkd * (sp_step - vc_est)
        ctrl_out = Ckd * state_k0 + Dkd * (sp_step - vc_est)

        # calculate the final drive to the cavity (slightly different from Fig.4.6, 
        # we do not divide controller output by b0)
        ctrl_step = ctrl_out - f / b0 + ff_step

    # return the results of the step
    return True, ctrl_step, ctrl_out, state_k, state_ob, vc_est, f

def AFF_timerev_lpf(vfb, fcut, fs, vff_cor = None):
    '''
    Time-reversed low-pass filter, we only apply the first order IIR low-pass,
    which gives up to 90 degrees phase lead.
    
    Refer to LLRF Book section 4.5.1.

    Parameters:
        vfb:     numpy array (complex), feedback control waveform
        fcut:    float, cut-off frequency of the low-pass filter, Hz
        fs:      float, sampling frequency, Hz
        vff_cor: numpy array (complex), buffer storing the filtered waveform
        
    Returns:
        status:  boolean, success (True) or fail (False)
        vff_cor: numpy array (complex), feedforward correction waveform
    '''
    # check the input
    if (vfb.shape[0] < 3) or (fcut <= 0.0) or \
       (fcut >= fs/2) or (fs <= 0.0):
        return False, None

    if vff_cor is not None:
        if (not vfb.shape == vff_cor.shape):
            return false, None

    # perform the time-reversed filtering 
    vfb = vfb[::-1]                         # reverse the time of input
    a   = 2.0 * np.pi * fcut / fs           # scale factor
    if vff_cor is None:
        vff_cor = np.zeros(vfb.shape, dtype = complex) # create buffer

    for i in range(1, vfb.shape[0]):        # filtering
        vff_cor[i] = (1.0 - a) * vff_cor[i-1] + a * vfb[i]

    # return the result
    return True, vff_cor[::-1]
    
def AFF_ilc_design(h, pulw, P = None, Q = None):
    '''
    Adaptive feedforward with optimal iterative learning control (ILC).
    
    Refer to LLRF Book section 4.5.2.

    Parameters:
        h:      numpy array (complex), impulse response
        pulw:   int, pulse width as number of points
        P, Q:   numpy matrix, positive-definite weight matrices
        
    Returns:
        status: boolean, success (True) or fail (False)
        L:      numpy matrix (complex), gain matrix of ILC
    '''
    # check the input
    if (h.shape[0] < 3) or (pulw < 3):
        return False, None

    if P is None:
        P = np.matrix(np.eye(pulw))
    if Q is None:
        Q = np.matrix(np.eye(pulw))

    if (not P.shape == Q.shape) or (not P.shape[0] == pulw):
        return False, None

    # dimension
    order = h.shape[0]

    # derive the system transfer matrix
    G = np.zeros((pulw, pulw), dtype = complex)
    for i in range(pulw):
        G[i:, i] = h[:min(pulw-i, order)]

    # calculate the ILC gain matrix
    G = np.matrix(G)
    L = np.linalg.inv(Q + G.H * P * G) * G.H * P

    return True, L

def AFF_ilc(vc_err, L):
    '''
    Apply the ILC.

    Refer to LLRF Book section 4.5.2.
    
    Parameters:
        vc_err:  numpy array (complex), error of the cavity voltage waveform
        L:       numpy matrix (complex), gain matrix of ILC
        
    Returns:
        vff_cor: numpy array (complex), feedforward correction waveform
    '''
    return np.matmul(L, vc_err)

def resp_inv_svd(R, singular_val_filt = 0.0):
    '''
    Response matrix inversion with SVD.

    Refer to Beam Control Book section 2.4.2.
    
    Parameters:
        R:                 numpy matrix, response matrix
        singular_val_filt: float, threshold of singular values, the ones 
                            smaller or equal to it will be discarded
                            
    Returns:
        Rinv: numpy matrix, inversion of the response matrix
    '''
    return np.linalg.pinv(R, rcond = singular_val_filt)

def resp_inv_lsm(R, regu = 0.0):
    '''
    Response matrix inversion with least-square method.
    
    Refer to Beam Control Book section 2.4.3.
    
    Parameters:
        R:    numpy matrix, response matrix
        regu: float, regularization factor (should > 0)
        
    Returns:
        Rinv: numpy matrix, inversion of the response matrix
    '''
    return np.linalg.inv(R.T * R + regu * np.matrix(np.eye(R.shape[1]))) * R.T

def impulse_resp_matrix(h, pulw):
    '''
    Calculate the response matrix for a LTI system with impulse response.    
    Refer to LLRF Book section 4.5.2.
    Parameters:
        h:      numpy array (complex), impulse response
        pulw:   int, pulse width as number of points       
    Returns:
        status: boolean, success (True) or fail (False)
        G:      numpy matrix (real), response matrix
    '''
    # check the input
    if (h.shape[0] < 3) or (pulw < 3):
        return False, None

    # dimension
    order = h.shape[0]

    # derive the system transfer matrix
    G = np.zeros((pulw, pulw), dtype = complex)
    for i in range(pulw):
        G[i:, i] = h[:min(pulw-i, order)]

    return True, G













"""Measure RF amplitude and phase from ADC samples."""
#############################################################################
#  Copyright (c) 2023 by Paul Scherrer Institute, Switzerland
#  All rights reserved.
#  Authors: Zheqiao Geng
#############################################################################
'''
#########################################################################
Here collects routines for RF signal detection and actuation

Assumptions:
  a. If not explicitly mentioned, all phases are in degree
 
Implemented:
    - noniq_demod   : perform non-I/Q demodulation of a given raw sampling waveform
    - twop_demod    : demodulate raw with every two samples
    - asyn_demod    : demodulate raw sampled by asyn. clock, reference WF needed
    - self_demod_ap : demodulate raw with Hilbert transform, return amplitude and phase
    - iq2ap_wf      : convert I/Q waveforms to amplitude/phase waveforms
    - ap2iq_wf      : convert amplitude/phase waveforms to I/Q waveforms
    - norm_phase    : normalize phase (scalar or WF) to a specific range (default +-180 deg)
    - pulse_info    : derive the pulse info like pulse width, pulse offset, etc.

Some algorithms are referred to the following books:
S. Simrock and Z. Geng, Low-Level Radio Frequency Systems, Springer, 2022
https://link.springer.com/book/10.1007/978-3-030-94419-3 ("LLRF Book")
#########################################################################
'''
import numpy as np
from scipy import signal

def noniq_demod(raw_wf, n, m = 1):
    '''
    Non-I/Q demodulation.

    Refer to LLRF Book section 5.2.2.
    
    Parameters:
        raw_wf: numpy array, 1-D array of the raw waveform
        m, n:   integer, non-I/Q parameters (n samples cover m IF cycles)
        
    Returns:
        status: boolean, success (True) or faile (False)
        I, Q:   numpy array, I/Q waveforms
    '''
    # check the input
    L = raw_wf.shape[0]
    if n <= 0 or m <= 0 or n <= m or L < n:
        return False, None, None

    # calculate the demod coefficients
    P_rad   = np.arange(0.0, n, 1) * 2.0 * m * np.pi / n    # phases of one superperiod of NCO
    I_coef  = np.sin(P_rad) * 2.0 / n                       # NCO output I
    Q_coef  = np.cos(P_rad) * 2.0 / n                       # NCO output Q

    # define the databuffers
    I       = np.zeros(L)       # final output buffer
    Q       = np.zeros(L)
    pipeI   = np.zeros(n + 1)   # FIFO helping calculation
    pipeQ   = np.zeros(n + 1)
    sumI    = 0.0               # accumulators
    sumQ    = 0.0
    coef_id = 0                 # index for data selection
    pipe_id = n

    # do demodulation
    for i in range(L):
        # fill the FIFO a new multiplication
        pipeI[pipe_id] = raw_wf[i] * I_coef[coef_id]
        pipeQ[pipe_id] = raw_wf[i] * Q_coef[coef_id]

        # calculate difference btw FIFO new/old, and accumulate
        sumI += pipeI[pipe_id] - pipeI[pipe_id-1]
        sumQ += pipeQ[pipe_id] - pipeQ[pipe_id-1]

        # update the output for an sample
        I[i] = sumI
        Q[i] = sumQ

        # update the index
        coef_id += 1
        if coef_id >= n:
            coef_id = 0

        pipe_id -= 1
        if pipe_id < 0:
            pipe_id = n

    # return the results
    return True, I, Q

def twop_demod(raw_wf, f_if, fs):
    '''
    Demodulation with two points.
    
    Refer to LLRF Book section 5.2.2.

    Parameters:
        raw_wf: numpy array, raw waveform to be demodulated
        f_if:   float, IF frequency, Hz
        fs:     float, sampling frequency, Hz
        
    Returns:
        status: boolean, success (True) or faile (False)
        I, Q:   numpy array, I/Q waveforms
    '''
    # check the input
    if (raw_wf.shape[0] < 3) or (f_if <= 0.0) or (fs <= 0.0):
        return False, None, None

    # reserve the buffer
    N = raw_wf.shape[0]                 # number of the points in WF
    I = np.zeros(N)
    Q = np.zeros(N)

    # make the demodulation    
    dphi_rad = 2.0 * np.pi * f_if / fs  # phase advance per sample
    sn_dphi  = np.sin(dphi_rad)    

    for i in range(1, N):
        I[i] = ( raw_wf[i] * np.cos((i-1)*dphi_rad) - raw_wf[i-1] * np.cos(i*dphi_rad)) / sn_dphi
        Q[i] = (-raw_wf[i] * np.sin((i-1)*dphi_rad) + raw_wf[i-1] * np.sin(i*dphi_rad)) / sn_dphi

    return True, I, Q

def asyn_demod(raw_wf, ref_wf):
    '''
    Asynchronous demodulation (need reference).

    Parameters:
        raw_wf:      numpy array, signal waveform to be demodulated
        ref_wf:      numpy array, samples of the RF reference signal
    Returns:
        status:      boolean, success (True) or faile (False)
        I, Q:        numpy array, I/Q waveforms of final demodulation
        Isig, Qsig:  numpy array, I/Q waveforms of raw_wf (with inaccurate phase)
        Iref, Qsig:  numpy array, I/Q waveforms of ref_wf (with inaccurate phase)
    '''
    # check the input
    if (not raw_wf.shape == ref_wf.shape) or (raw_wf.shape[0] < 3):
        return (False,) + (None,)*6

    # estimate the reference frequency 
    N = raw_wf.shape[0]                         # number of the points in WF
    Y = np.fft.fft(ref_wf)
    f_if = np.argmax(np.abs(Y[:round(N/2)]))    # find the peak, respresenting the IF freq
    fs   = N                                    # represent the sampling frequency

    # demodulate with twop_demod
    status1, Isig, Qsig = twop_demod(raw_wf, f_if, fs)
    status2, Iref, Qref = twop_demod(ref_wf, f_if, fs)
    if not (status1 and status2):
        return (False,) + (None,)*6

    # get the reference phase
    status, Aref, Pref_deg = iq2ap_wf(Iref, Qref)
    if not status:
        return (False,) + (None,)*6

    # unwrap the reference phase
    Pref_deg = np.unwrap(Pref_deg, period = 360.0)

    # get the linear fitting of the phase (exclude first points)
    T = np.arange(N)   
    p = np.polyfit(T[2:], Pref_deg[2:], 1)

    # reconstruct the clean phase of reference
    Pref_rec_deg  = np.polyval(p, T)
    Pref_rec_deg -= Pref_rec_deg[0]

    # remove the phase slope from the signal phase
    C = (Isig + 1j*Qsig) * np.exp(-1j * Pref_rec_deg * np.pi / 180.0)

    return True, np.real(C), np.imag(C), Isig, Qsig, Iref, Qref

def self_demod_ap(raw_wf, n = 1):
    '''
    Self demodulation with Hilbert transform (later may implement padding to
    mitigate the edge effects).

    Parameters:
        raw_wf: numpy array, signal waveform to be demodulated
        n:      int, number of points covering full cycles (for coherent sampling)
    Returns:
        status: boolean, success (True) or faile (False)
        A, P:   numpy array, amplitude and phase waveforms, P in degree
    '''
    # check the input
    if raw_wf.shape[0] < 3:
        return False, None, None

    # perform Hilbert transform
    N = raw_wf.shape[0]             # number of points in WF
    L = int(N / n) * n              # tailor
    C = signal.hilbert(raw_wf[:L])  # hilbert transform
    
    # to the correct length
    if N > L:
        C = np.hstack((C, np.zeros(N - L)))

    # get the amplitude and phase in deg
    A = np.abs(C)
    P = np.angle(C, deg = True)

    return True, A, P

def iq2ap_wf(I, Q):
    '''
    I/Q to A/P waveforms.
    
    Parameters:
        I, Q: numpy array, I/Q waveforms
        
    Returns:
        A, P: numpy array, amplitude/phase waveforms, P in degree
    '''
    # check the input
    if not I.shape == Q.shape:
        return False, None, None

    # convert to complex
    C = I + 1j*Q

    # return the results
    return True, np.abs(C), np.angle(C, deg = True)

def ap2iq_wf(A, P):
    '''
    A/P to I/Q waveforms.
    
    Parameters:
        A, P: numpy array, amplitude/phase waveforms, P in degree
        
    Returns:
        I, Q: numpy array, I/Q waveforms
    '''
    # check the input
    if not A.shape == P.shape:
        return False, None, None

    # do calculation
    P_rad = P * np.pi / 180.0
    I = A * np.cos(P_rad)
    Q = A * np.sin(P_rad)

    # return the results
    return True, I, Q

def norm_phase(P, cmd = '+-180'):
    '''
    normalize phase to +-180 degree or between 0 to 360 degree.
    
    Parameters:
        P:     float, phase in degree
        cmd:   string '+-180' or '0to360'
    Returns:
        Pnorm: float, normalized phase, degree
    '''
    # convert using the complex numbers
    Pnorm = np.angle(np.exp(1j * P * np.pi / 180.0), deg = True)    # to +-180 deg
    if cmd == '0to360':
        if isinstance(Pnorm, float):
            Pnorm = (Pnorm + 360.0) if (Pnorm < 0.0) else Pnorm
        else:
            Pnorm[Pnorm < 0.0] += 360

    # return the result
    return Pnorm

def pulse_info(pulse, threshold = 0.1):
    '''
    get information of a pulse.

    Parameters:
        pulse:     numpy array, the pulse data
        threshold: float, threshold for edge detection 
    Returns:
        offs:      int, offset id of the pulse
        pulw:      int, pulse width as number of samples
        peak:      float, peak magnitude of the pulse

    To be done:
        a. detect the rise time and falling time
        b. detect the flattop region and average
        c. calculate the energy in the pulse
    '''
    result = {'status': False}

    # check the input
    if isinstance(pulse, list):
        data = np.array(pulse)
    elif isinstance(pulse, np.ndarray):
        data = pulse
    else:
        return result

    if data.shape[0] <= 1 or threshold < 0.0:
        return result

    # be sure the pulse is positive
    data = np.abs(data)

    # get the peak
    result['peak'] = np.max(data)           # peak value of the pulse
    low_l = threshold * result['peak']      # the low limit of the pulse

    # get the pulse info
    over_lowl = data > low_l
    n         = over_lowl.shape[0]          # size of the pulse
    start_id  = 0                           # init the pulse start id
    end_id    = n - 1                       # init the pulse end id

    for i in range(2, n - 1):
        # find the pulse start id
        if (not any(over_lowl[:i])) and all(over_lowl[i:i+2]):
            start_id = i
    
        # find the pulse end id
        if (not any(over_lowl[i:])) and all(over_lowl[i-2:i]):
            end_id = i - 1

    result['offs']   = start_id
    result['pulw']   = end_id - start_id + 1
    result['status'] = True

    # return the results
    return result
































"""Simulate the RF cavity response in the presence of RF drive and beam loading."""
#############################################################################
#  Copyright (c) 2023 by Paul Scherrer Institute, Switzerland
#  All rights reserved.
#  Authors: Zheqiao Geng
#############################################################################
'''
#########################################################################
Here collects routines for RF system simulator

Implemented:
    - cav_ss                : derive a continous state-space equation of a cavity (all modes)
    - cav_ss_passband       : derive a continous state-space equation of a cavity (only passband modes)
    - cav_ss_mech           : derive a continous state-space equation of mechanical modes
    - cav_impulse           : derive the cavity impulse response from the cavity parameters
    - sim_ncav_pulse        : simulate cavity (with constant QL and detuning) response to a pulsed input
    - sim_ncav_step         : simulate cavity (with constant QL and detuning) response for a time step
    - sim_ncav_step_simple  : simulate cavity (with constant QL and detuning) response for a time step
                              (simplified cavity equation only with the fundamental passband mode)
    - sim_scav_step         : simulate cavity response with mechanical modes for a time step
    - sim_ss_step           : a generic state-space solver to execute for one step
    - rf_power_req          : calculate the required RF power for diesired cavity voltage and beam
    - opt_QL_detuning       : calcualte the optimal QL and detuning for minimizing the reflection power

To be implemented:
    - implement the superconducting cavity model with Lorenz force detuning
    - traveling-wave structure model
    - pulse compressor (SLED/BOC) model
 
Some algorithms are referred to the following books:
S. Simrock and Z. Geng, Low-Level Radio Frequency Systems, Springer, 2022
https://link.springer.com/book/10.1007/978-3-030-94419-3 ("LLRF Book")
#########################################################################
'''
import numpy as np
from scipy import signal

from llrflibs.rf_sysid import *
from llrflibs.rf_misc import *

def cav_ss(half_bw, detuning = 0.0, beta = 1e4, passband_modes = None, 
           plot = False, plot_pno = 1000):
    '''
    Derive the continuous state-space equation of the cavity
     - include pass-band modes.
     - we assume the half-bandwidth and detuning of the fundamental mode are constant.
     - we output two models: one for RF drive, and the other for beam (beam only 
       interacts with the fundamental mode of the cavity).
       
    Refer to LLRF Book section 3.3.7 and 3.4.3.
    
    Parameters:
        half_bw:        float, half bandwidth of the cavity (constant), rad/s
        detuning:       float, detuning of the cavity (constant), rad/s
        beta:           float, input coupling factor (needed for NC cavities; 
                         for SC cavities, can use the default value, or you can 
                         specify it if more accurate result is needed)
        passband_modes: dict, with the following items:
                        ``freq_offs`` : list, offset frequencies of the modes, Hz;
                        ``gain_rel``  : list, relative gain wrt fundamental mode;
                        ``half_bw``   : list, half bandwidth of the mode, rad/s
        plot:           boolean, enable the plot of frequency response
        plot_pno:       int, number of point in the plot  
        
    Returns:
        status:             boolean, success (True) or fail (False)
        Arf, Brf, Crf, Drf: numpy matrix (complex), continous cavity model for RF drive
        Abm, Bbm, Cbm, Dbm: numpy matrix (complex), continous cavity model for beam drive
    '''
    # check the parameters
    if (half_bw <= 0) or (beta <= 0):
        return (False,) + (None,)*8

    if passband_modes is not None:
        if (not isinstance(passband_modes, dict)) or \
           ('freq_offs' not in passband_modes.keys()) or \
           ('gain_rel' not in passband_modes.keys()) or \
           ('half_bw' not in passband_modes.keys()): 
            return (False,) + (None,)*8

    plot_pno = 1000 if (plot_pno <= 0) else plot_pno

    # define the parameters of the fundamental mode
    b0 = 2 * beta * half_bw / (beta + 1)            # gain for the RF drive voltage
    b1 = 2 * half_bw                                # gain for the beam drive voltage

    # transfer function of the fundamental mode
    rf_num = [b0]
    rf_den = [1, half_bw - 1j*detuning]
    bm_num = [b1]
    bm_den = [1, half_bw - 1j*detuning]

    # interprete the passband modes
    if passband_modes is not None:
        # get the passband mode parameters
        pb_f  = passband_modes['freq_offs']         # frequency offset compared to the fundamental mode, Hz
        pb_g  = passband_modes['gain_rel']          # gain relative to the fundamental mode
        pb_wh = passband_modes['half_bw']           # half bandwidth of the passband modes, rad/s

        # add the passband mode transfer functions
        for i in range(len(pb_f)):
            rf_num, rf_den = add_tf(rf_num, rf_den, 
                                    [b0 * pb_g[i] * pb_wh[i] / half_bw], 
                                    [1, pb_wh[i] - 1j*2*np.pi*pb_f[i]])

    # get the state-space model
    Arf, Brf, Crf, Drf = signal.tf2ss(rf_num, rf_den)
    Abm, Bbm, Cbm, Dbm = signal.tf2ss(bm_num, bm_den)

    # plot the response
    if plot:
        # get the max freq range
        if passband_modes is not None:
            max_wrange = np.max([abs(detuning), 
                                 5*half_bw,
                                 np.max(np.abs(passband_modes['freq_offs']))*2*np.pi])
        else:
            max_wrange = np.max([abs(detuning), 
                                 5*half_bw])

        # calculate the response
        wrf, hrf = signal.freqs(rf_num, rf_den, worN = np.linspace(-2*max_wrange, 2*max_wrange, plot_pno))
        wbm, hbm = signal.freqs(bm_num, bm_den, worN = np.linspace(-2*max_wrange, 2*max_wrange, plot_pno))

        # make the plot
        from llrflibs.rf_plot import plot_cav_ss
        plot_cav_ss(wrf, hrf, wbm, hbm)

    # return the results
    return True, Arf, Brf, Crf, Drf, Abm, Bbm, Cbm, Dbm

def cav_ss_passband(passband_modes):
    '''
    Derive the continuous state-space equation of the cavity, only the passband modes.      
    Refer to LLRF Book section 3.3.7 and 3.4.3.
    
    Parameters:
        passband_modes: dict, with the following items:
                        ``freq_offs`` : list, offset frequencies of the modes, Hz;
                        ``gain_rel``  : list, relative gain wrt fundamental mode;
                        ``half_bw``   : list, half bandwidth of the mode, rad/s
    Returns:
        status:             boolean, success (True) or fail (False)
        Arf, Brf, Crf, Drf: numpy matrix (complex), continous passband model for RF drive
    '''
    # check the parameters
    if passband_modes is None:
        return (False,) + (None,)*4

    if (not isinstance(passband_modes, dict)) or \
       ('freq_offs' not in passband_modes.keys()) or \
       ('gain_rel'  not in passband_modes.keys()) or \
       ('half_bw'   not in passband_modes.keys()): 
        return (False,) + (None,)*4

    if len(passband_modes['freq_offs']) < 1:
        return (False,) + (None,)*4       

    # transfer function initialization
    rf_num = [0.0]
    rf_den = [1.0]

    # get the passband mode parameters
    pb_f  = passband_modes['freq_offs']         # frequency offset compared to the fundamental mode, Hz
    pb_g  = passband_modes['gain_rel']          # gain relative to the fundamental mode
    pb_wh = passband_modes['half_bw']           # half bandwidth of the passband modes, rad/s

    # add the passband mode transfer functions
    for i in range(len(pb_f)):
        rf_num, rf_den = add_tf(rf_num, rf_den, 
                                [pb_g[i] * pb_wh[i]], 
                                [1, pb_wh[i] - 1j*2*np.pi*pb_f[i]])

    # get the state-space model
    Arf, Brf, Crf, Drf = signal.tf2ss(rf_num, rf_den)

    # return the results
    return True, Arf, Brf, Crf, Drf

def cav_ss_mech(mech_modes, lpf_fc = None):
    '''
    Derive the continous state-space equation of the cavity mechanical modes.
    
    Refer to LLRF book section 3.3.10.
    
    Parameters:
        mech_modes: dict, with the following items:
            ``f`` : list, frequencies of mech modes, Hz;
            ``Q`` : list, qualify factors;
            ``K`` : list, K values, rad/s/(MV)^2.
        lpf_fc    : float, low-pass cutoff freq (optional), Hz
    Returns:
        status:     boolean, success (True) or fail (False)
        A, B, C, D: numpy matrix (real), continous mechanical model    
    '''
    # check 
    if mech_modes is None:
        return (False,) + (None,)*4
    
    if len(mech_modes['f']) < 1:
        return (False,) + (None,)*4
    
    # get the paremters
    fm = mech_modes['f']                # frequency of the mechanical mode, Hz
    Qm = mech_modes['Q']                # quality factor
    Km = mech_modes['K']                # K value, rad/s/(MV)^2    
    
    # build the transfer function of the first mechanical mode
    w = 2 * np.pi * fm[0]               # get the angular freq, rad/s
    num = [-Km[0] * w**2]
    den = [1, w/Qm[0], w**2]
        
    # add the other mode transfer functions
    if len(fm) > 1:
        for i in range(1, len(fm)):
            w = 2 * np.pi * fm[i]
            num, den = add_tf(num, den, [-Km[i] * w**2], [1, w/Qm[i], w**2])

    # add the LPF if applicable
    if lpf_fc is not None:
        if lpf_fc > 0.0:
            w = 2 * np.pi * lpf_fc
            num, den = add_tf(num, den, [w], [1, w])

    # get the state-space model
    A, B, C, D = signal.tf2ss(num, den)    
    
    # return the results
    return True, A, B, C, D

def cav_impulse(half_bw, detuning, Ts, order = 20):
    '''
    Derive the impulse response from the cavity equation. We assume that the 
    system gain has been normalized to 1 and the system phase corrected to 0. 
    Therefore, the referred cavity equation is ``dvc/dt + (half_bw - 1j * detuning)*vc = half_bw * vd``, 
    where ``vc`` is the vector of cavity voltage and ``vd`` is the vector of cavity 
    drive (in principle ``vd = 2 * beta * vf / (beta + 1)``).

    Refer to LLRF Book section 4.5.2.

    Parameters:
        half_bw:  float, constant half bandwidth of the cavity, rad/s
        detuning: float, constant detuning of the cavity, rad/s
        Ts:       float, sampling time, s
        order:    int, order of the impulse response    
        
    Returns:
        status:   boolean, success (True) or fail (False)
        h:        numpy array (complex), impulse response
    '''
    # check the input
    if (half_bw <= 0.0) or (Ts <= 0.0) or \
       (order < 2):
        return False, None

    # calculate the impulse response
    k = np.arange(order)
    h = Ts * half_bw * (1.0 - Ts * (half_bw - 1j * detuning))**k

    # return the results
    return True, h
    
def boc_impulse(half_bw, detuning, Ts, order = 20, beta = 6):
    '''
    Derive the BOC impulse response using the cav_impulse function.

    Parameters:
        half_bw:  float, constant half bandwidth of the cavity, rad/s
        detuning: float, constant detuning of the cavity, rad/s
        Ts:       float, sampling time, s
        order:    int, order of the impulse response    
        beta:     float, input coupling factor of BOC
        
    Returns:
        status:   boolean, success (True) or fail (False)
        h:        numpy array (complex), impulse response
    '''
    # check the input
    if (half_bw <= 0.0) or (Ts <= 0.0) or (beta <= 0.0) or \
       (order < 2):
        return False, None

    # get the cavity response
    status, h_cav = cav_impulse(half_bw, detuning, Ts, order = order)
    if not status:
        return False, None
        
    # derive the BOC impulse response
    h = h_cav * 2.0 * beta / (beta + 1.0)
    h[0] -= 1.0

    # return the results
    return True, h       

def sim_ncav_pulse(Arfc, Brfc, Crfc, Drfc, vf, Ts, 
                   Abmc = None, 
                   Bbmc = None, 
                   Cbmc = None, 
                   Dbmc = None, 
                   vb   = None):
    '''
    Simulate the cavity response to a pulsed RF drive and beam current. This
    function is for normal conducting cavties with constant QL and detuning.

    Parameters:
        Arfc, Brfc, Crfc, Drfc: numpy matrix (complex), continous cavity model for RF drive
        vf:                     numpy array (complex), cavity forward voltage (calibrated to
                                 the cavity probe signal reference plane)
        Ts:                     float, sampling frequency, Hz
        Abmc, Bbmc, Cbmc, Dbmc: numpy matrix (complex), continous cavity model for beam drive
        vb:                     numpy array (complex), beam drive voltage (calibrated to
                                 the cavity probe signal reference plane)
                                 
    Returns:
        status: boolean, success (True) or fail (False)
        T:      numpy array, time waveform, s
        vc:     numpy array (complex), cavity voltage waveform
        vr:     numpy array (complex), cavity reflected voltage waveform
    '''
    # check the input
    if (Ts <= 0.0):
        return False, None, None, None

    if vb is not None:
        if (not vb.shape == vf.shape):
            return False, None, None, None

    # simulate the response of the continous system
    T = np.arange(vf.shape[0]) * Ts
    _, vc_rf, _ = signal.lsim((Arfc, Brfc, Crfc, Drfc), vf, T)      # Returns: T, Yout, Xout
    vc = vc_rf
    if not any([x is None for x in (Abmc, Bbmc, Cbmc, Dbmc, vb)]):
        _, vc_bm, _ = signal.lsim((Abmc, Bbmc, Cbmc, Dbmc), vb, T)
        vc += vc_bm

    # get the cavity reflected
    vr = vc - vf

    return True, T, vc, vr

def sim_ncav_step(Arfd, Brfd, Crfd, Drfd, vf_step, state_rf0,
                  Abmd      = None, 
                  Bbmd      = None, 
                  Cbmd      = None, 
                  Dbmd      = None, 
                  vb_step   = None,  
                  state_bm0 = None):
    '''
    Simulate the cavity response for a time step using the discrete cavity 
    state-space function.

    Parameters:
        Arfd, Brfd, Crfd, Drfd: numpy matrix (complex), discrete cavity model for RF drive
        vf_step:                complex, cavity forward voltage of this step
        state_rf0:              numpy matrix (complex), state of RF response model of last step
        Abmd, Bbmd, Cbmd, Dbmd: numpy matrix (complex), discrete cavity model for beam drive
        vb_step:                complex, beam drive voltage of this step
        state_bm0:              numpy matrix (complex), state of beam response model of last step
        
    Returns:
        status:   boolean, success (True) or fail (False)
        vc_step:  complex, cavity voltage of this step
        vr_step:  complex, cavity reflected voltage of this step
        state_rf: numpy matrix (complex), state of RF response model of this step (should input
                   to next execution)
        state_bm: numpy matrix (complex), state of beam response model of this step (should input
                   to next execution)
    '''
    # calculate the RF/beam response
    state_rf    = Arfd * state_rf0 + Brfd * vf_step
    vc_rf_step  = Crfd * state_rf0 + Drfd * vf_step

    if not any([x is None for x in (Abmd, Bbmd, Cbmd, Dbmd, vb_step, state_bm0)]):
        state_bm    = Abmd * state_bm0 + Bbmd * vb_step
        vc_bm_step  = Cbmd * state_bm0 + Dbmd * vb_step
        vc_step     = vc_rf_step[0,0] + vc_bm_step[0,0]
    else:
        state_bm    = None
        vc_step     = vc_rf_step[0,0]

    # get the cavity reflected of this step
    vr_step = vc_step - vf_step

    # return the results of the step
    return True, vc_step, vr_step, state_rf, state_bm

def sim_ncav_step_simple(half_bw, detuning, vf_step, vb_step, vc_step0, Ts, beta = 1e4):
    '''
    Simulate the cavity response for a time step using the simple discrete
    cavtiy equation (Euler method for discretization).

    Parameters:
        half_bw:  float, half bandwidth of the cavity (constant), rad/s
        detuning: float, detuning of the cavity (constant), rad/s
        vf_step:  complex, cavity forward voltage of this step
        vb_step:  complex, beam drive voltage of this step
        vc_step0: complex, cavity voltage of the last step
        Ts:       float, sampling time, s
        beta:     float, input coupling factor (needed for NC cavities; 
                   for SC cavities, can use the default value, or you can 
                   specify it if more accurate result is needed)   
    Returns:
        status:   boolean, success (True) or fail (False)
        vc_step:  complex, cavity voltage of this step
        vr_step:  complex, cavity reflected voltage of this step
    '''
    # check the input
    if (half_bw <= 0.0) or (Ts <= 0.0) or (beta <= 0.0):
        return False, None, None

    # make a step of calculation
    vc_step = (1 - Ts * (half_bw - 1j*detuning)) * vc_step0 + \
              2 * half_bw * Ts * (beta * vf_step / (beta + 1) + vb_step)
    vr_step = vc_step - vf_step

    # return the results of the step
    return True, vc_step, vr_step

def sim_scav_step(half_bw, dw_step0, detuning0, vf_step, vb_step, vc_step0, Ts, beta = 1e4,
                  state_m0 = 0, Am = None, Bm = None, Cm = None, Dm = None, mech_exe = False):
    '''
    Simulate the cavity response for a time step using the simple discrete
    cavtiy equation (Euler method for discretization) including the mechanical
    modes. We first simulate one step of the mechanical mode and determine the 
    detuning value, then use the ``sim_ncav_step_simple`` function to simulate
    one step of the electrical model.

    Parameters:
        half_bw:   float, half bandwidth of the cavity (constant), rad/s
        dw_step0:  float, detuning of the last step, rad/s
        detuning0: float, external detuning (tuner + microphonics), rad/s
        vf_step:   complex, cavity forward voltage of this step, V
        vb_step:   complex, beam drive voltage of this step, V
        vc_step0:  complex, cavity voltage of the last step, V
        Ts:        float, sampling time, s
        beta:      float, input coupling factor (needed for NC cavities; 
                    for SC cavities, can use the default value, or you can 
                    specify it if more accurate result is needed)   
        state_m0:  numpy matrix (real), last state of the mechanical equation 
        Am, Bm, Cm, Dm: numpy matrix (real), state-space matrix of mech modes
        mech_exe:  boolean, if exe mech sim one step or not (for down sampling)
    Returns:
        status:   boolean, success (True) or fail (False)
        vc_step:  complex, cavity voltage of this step
        vr_step:  complex, cavity reflected voltage of this step
        dw:       float, detuning of this step, rad/s
        state_m:  numpy matrix (real), updated state of the mechanical equation
    '''
    # check the input
    if (half_bw <= 0.0) or (Ts <= 0.0) or (beta <= 0.0):
        return (False,) + (None,)*4

    # make a step of calculation of electrical equation (only pi mode)
    vc_step = (1 - Ts * (half_bw - 1j*dw_step0)) * vc_step0 + \
              2 * half_bw * Ts * (beta * vf_step / (beta + 1) + vb_step)
    vr_step = vc_step - vf_step

    # update the mechanical mode equation and get the detuning    
    if (state_m0 is None) or \
       (Am is None) or \
       (Bm is None) or \
       (Cm is None) or \
       (Dm is None) or \
       (not mech_exe):
        state_m = state_m0
        dw      = detuning0
    else:
        state_m = Am * state_m0 + Bm * (abs(vc_step) * 1.0e-6)**2
        dw      = Cm * state_m0 + Dm * (abs(vc_step) * 1.0e-6)**2 + detuning0
       
    # return the results of the step
    return True, vc_step, vr_step, dw, state_m

def sim_ss_step(Ad, Bd, Cd, Dd, vin_step, state0):
    '''
    A generic state-space solver to execute for one step.
    Parameters:
        Ad, Bd, Cd, Dd: numpy matrix (float/complex), discrete state-space matrices
        vin_step:   float/complex, input of this step
        state0:     numpy matrix (float/complex), state of last step
    Returns:
        status:     boolean, success (True) or fail (False)
        vout_step:  float/complex, output of this step
        state:      numpy matrix (float/complex), state of this step (input to next exe)
    '''
    state     = Ad * state0 + Bd * vin_step
    vout_step = Cd * state0 + Dd * vin_step
    return True, vout_step, state

def rf_power_req(f0, vc0, ib0, phib, Q0, roQ_or_RoQ, 
                 QL_vec       = None,
                 detuning_vec = None, 
                 machine      = 'linac',
                 plot         = False):
    '''
    Plot the steady-state forward and reflected power for given cavity voltage, 
    beam current and beam phase, as function of loaded Q and detuning. The beam 
    phase is defined to be zero for on-crest acceleration.

    Refer to LLRF Book section 3.3.9.

    Parameters:
        f0:           float, RF operating frequency, Hz
        vc0:          float, desired cavity voltage, V
        ib0:          float, desired average beam current, A
        phib:         float, desired beam phase, degree
        Q0:           float, unloaded quality factor (for SC cavity, give it a
                       very high value like 1e10)
        roQ_or_RoQ:   float, cavity r/Q of Linac or R/Q of circular accelerator, Ohm
                       (see the note below)
        QL_vec:       numpy array, QL vector for power calculation
        detuning_vec: numpy array, detuning at which to evaluated the powers
        machine:      string, ``linac`` or ``circular``, used to select r/Q or R/Q
        plot:         boolean, enable/disable the plotting
        
    Returns:
        status:       boolean, success (True) or fail (False)
        Pfor:         dictionary, keyed by detuning, forward power at different QL
        Pref:         dictionary, keyed by detuning, reflected power at different QL

    Note: 
          Linacs define the ``r/Q = Vacc**2 / (w0 * U)`` while circular machines 
          define ``R/Q = Vacc**2 / (2 * w0 * U)``, where ``Vacc`` is the accelerating
          voltage, ``w0`` is the angular cavity resonance frequency and ``U`` is the
          cavity energy storage. Therefore, to use this function, one needs to 
          specify the ``machine`` to be ``linac`` or ``circular``. Generally, we have
          ``R/Q = 1/2 * r/Q``.
    '''
    # check the input
    if (f0 <= 0.0) or (vc0 < 0.0) or (ib0 < 0.0) or (Q0 <= 0.0) or \
       (roQ_or_RoQ <= 0.0) or (QL_vec.shape[0] < 1):
        return False, None, None

    if detuning_vec is None:
        detuning_vec = [0.0]

    # calcualte the necessary parameters
    if machine == 'circular': RL_vec = roQ_or_RoQ * QL_vec          # calculate the loaded resistance RL, Ohm
    else:                     RL_vec = 0.5 * roQ_or_RoQ * QL_vec
    beta_vec = Q0 / QL_vec - 1.0                                    # input coupling factor
    wh_vec   = np.pi * f0 / QL_vec                                  # hald bandwidth, rad/s
    phib_rad = phib * np.pi / 180.0                                 # beam phase in radian

    # calculate for each detuning
    Pfor = {}
    Pref = {}
    for dw in detuning_vec:
        Pfor[dw] = (beta_vec + 1) / beta_vec * vc0**2 / 8 / RL_vec * (\
                   (1 + 2 * RL_vec * ib0 * np.cos(phib_rad) / vc0)**2 + \
                   (dw / wh_vec + 2 * RL_vec * ib0 * np.sin(phib_rad) / vc0)**2)
        Pref[dw] = (beta_vec + 1) / beta_vec * vc0**2 / 8 / RL_vec * (\
                   ((beta_vec - 1)/(beta_vec + 1) - 2 * RL_vec * ib0 * np.cos(phib_rad) / vc0)**2 + \
                   (dw / wh_vec + 2 * RL_vec * ib0 * np.sin(phib_rad) / vc0)**2)

    # plot the result
    if plot:
        from llrflibs.rf_plot import plot_rf_power_req
        plot_rf_power_req(Pfor, Pref, QL_vec)

    return True, Pfor, Pref

def opt_QL_detuning(f0, vc0, ib0, phib, Q0, roQ_or_RoQ, 
                    machine  = 'linac', 
                    cav_type = 'sc'):
    '''
    Derived the optimal loaded Q and detuning.

    Refer to LLRF Book section 3.3.9.

    Parameters:
        f0:           float, RF operating frequency, Hz
        vc0:          float, desired cavity voltage, V
        ib0:          float, desired average beam current, A
        phib:         float, desired beam phase, degree
        Q0:           float, unloaded quality factor (for SC cavity, 
                       give it a very high value like 1e10)
        roQ_or_RoQ:   float, cavity r/Q of Linac or R/Q of circular accelerator, Ohm
                       (see the note below)
        machine:      string, ``linac`` or ``circular``, used to select r/Q or R/Q
        cav_type:     string, ``sc`` for superconducting or ``nc`` for normal conducting
        
    Returns:
        status:       boolean, success (True) or fail (False)
        QL_opt:       float, optimal loaded quality factor
        dw_opt:       float, optimal detuning, rad/s
        beta_opt:     float, optimal input coupling factor

    Note: 
          Linacs define the ``r/Q = Vacc**2 / (w0 * U)`` while circular machines 
          define ``R/Q = Vacc**2 / (2 * w0 * U)``, where ``Vacc`` is the accelerating
          voltage, ``w0`` is the angular cavity resonance frequency and ``U`` is the
          cavity energy storage. Therefore, to use this function, one needs to 
          specify the ``machine`` to be ``linac`` or ``circular``. Generally, we have
          ``R/Q = 1/2 * r/Q``.
    '''
    # check the input
    if (f0 <= 0.0) or (vc0 < 0.0) or (ib0 < 0.0) or (Q0 <= 0.0) or \
       (roQ_or_RoQ <= 0.0):
        return False, None, None

    # some parameters
    if machine == 'circular': shunt_imp = roQ_or_RoQ * 2.0
    else:                     shunt_imp = roQ_or_RoQ
    phib_rad = phib * np.pi / 180.0                   # beam phase in radian

    # calculate the optimal values
    dw_opt = -np.pi * f0 * shunt_imp * ib0 * np.sin(phib_rad) / vc0
    
    if cav_type == 'sc':
        QL_opt   = vc0 /(shunt_imp * ib0 * np.cos(phib_rad))
        beta_opt = Q0 / QL_opt - 1.0
    else:
        beta_opt = shunt_imp * Q0 * ib0 * np.cos(phib_rad) / vc0 + 1.0
        QL_opt   = Q0 / (beta_opt + 1.0)

    # return the results
    return True, QL_opt, dw_opt, beta_opt





















